//
// Mengzi Zhang
// 24 Sept 2010
// CSE 167 hw 1
//

#include <Vector3.h>

// Default ctor
Vector3::Vector3 ()
{
  num_elts = 3;
  for (int i = 0; i < num_elts; i++)
    v[i] = 0;
}

// ctor
Vector3::Vector3 (double v0, double v1, double v2)
{
  num_elts = 3;
  v[0] = v0;
  v[1] = v1;
  v[2] = v2;
}

void Vector3::set (int idx, double val)
{
  if ((idx < 0) || (idx >= num_elts))
  {
    fprintf (stderr,
      "Error: Vector3::set(): index %d out of bounds, setting to -1\n", idx);
    v[idx] = -1;
  }

  v[idx] = val;
}

void Vector3::setAll (double v1, double v2, double v3)
{
  v[0] = v1;
  v[1] = v2;
  v[2] = v3;
}

double Vector3::get (int idx)
{
  if ((idx < 0) || (idx >= num_elts))
  {
    fprintf (stderr,
      "Error: Vector3::get(): index %d out of bounds, returning -1\n", idx);
    return -1;
  }

  return v[idx];
}

// Overload get() with operator []
double Vector3::operator[] (int idx)
{
  return get (idx);
}

// res = this_vec + vec2
void Vector3::add (Vector3& vec2, Vector3& res)
{
  for (int i = 0; i < num_elts; i++)
    res.set (i, v [i] + vec2 [i]);
}

// Potential bug: is this the right way to return res? Make the ret val by
//   reference? If use Vector3& as ret type, then would be returning a local
//   var, will be out of scope...
Vector3 Vector3::operator+ (Vector3& vec2)
{
  Vector3 res = Vector3 ();
  add (vec2, res);
  return res;
}

// res = this_vec - vec2
void Vector3::sub (Vector3& vec2, Vector3& res)
{
  for (int i = 0; i < num_elts; i++)
    res.set (i, v[i] - vec2[i]);
}

Vector3 Vector3::operator- (Vector3& vec2)
{
  Vector3 res = Vector3 ();
  sub (vec2, res);
  return res;
}

// res = scalar * this_vec
void Vector3::mul (double scalar, Vector3& res)
{
  for (int i = 0; i < num_elts; i++)
    res.set (i, scalar * v[i]);
}

// Dot product of vecs u=[a,b,c] and v=[d,e,f] is defined as:
//   a*d + b*e + c*f
// which may also be obtained by transposing first vec, uTv.
double Vector3::dot (Vector3& vec2)
{
  return (v[0] * vec2[0] + v[1] * vec2[1] + v[2] * vec2[2]);
}

// Cross product of vecs a=[a1,a2,a3] and b=[b1,b2,b3] is defined as:
//   a x b = (a2b3 - a3b2)i + (a3b1 - a1b3)j + (a1b2 - a2b1)k
//   or, in terms of determinants (easier to see)
//     |i  j  k |
//     |a1 a2 a3|
//     |b1 b2 b3|
//   or,
//   a x b = n a b sin q, q is angle btwn a and b, n is the Unit vector (mag
//     == 1)
void Vector3::cross (Vector3& vec2, Vector3& res)
{
  res.set (0, v[1]*vec2[2] - v[2]*vec2[1]);
  // Potential bug: shouldn't this be a1b3-a3b1, instead of the other way?
  // TODO: how do we test cross product?
  res.set (1, v[2]*vec2[0] - v[0]*vec2[2]);
  res.set (2, v[0]*vec2[1] - v[1]*vec2[0]);
}

// Length of this_vec
//   len = sqrt (v0^2 + v1^2 + v2^2)
double Vector3::len ()
{
  double v0 = v[0], v1 = v[1], v2 = v[2];
  return (sqrt (v0*v0 + v1*v1 + v2*v2));
}

// Returns normalized vec of this_vec
void Vector3::norm (Vector3& res)
{
  double length = len ();

  if (length != 0)
    for (int i = 0; i < num_elts; i++)
      res.set (i, v[i] / length);
}

// Returns true if vec2's all components are same as componenets of this_vec
bool Vector3::equals (Vector3& vec2)
{
  double v0 = v[0], v1 = v[1], v2 = v[2];
 
  // If any one component is not equal
  if ((v0 != vec2 [0]) || (v1 != vec2 [1]) || (v2 != vec2 [2]))
    return false;
  else
    return true;
}

void Vector3::print ()
{
  fprintf (stderr, "(%f, %f, %f)\n", v[0], v[1], v[2]);
}

