// 
// (Mabel) Mengzi Zhang
// 30 Oct 2010
// CSE 167 hw 5
//

#ifndef _SHAPE3D_H_
#define _SHAPE3D_H_

// scenegraph
#include <leaf.h>
#include <node.h>

#include <shape.h>
#include <Vector3.h>
#include <objReader.h>

#define SHAPE_SPHERE 1
#define SHAPE_CUBE 2
#define SHAPE_CONE 3
#define SHAPE_OBJ 4

// From hw 2
class ObjFile : public Shape
{
  public:

    ObjFile ();

    float min (float num1, float num2);

    void loadObj (float * minmax, Matrix4& transformMat, char * objFilename);
    void normalize (float * minmax, Matrix4& transformMat);
    void draw ();

  protected:

    // Vertices
    // v
    int nVerts;
    float * verts;
    // vn
    float * normals;
    // vt
    float * texcoords;
    // f. Index of 3 vertices that make up a triangle
    int nIndices;
    int * indices;
};


class Shape3D : public Leaf
{
  public:
    // Use virtual keyword to make sure the right destructor is called, i.e.
    //   TransformGroup's dtor, not Group (parent)'s detor.
    Shape3D ();
    Shape3D (Shape * s);
    Shape3D (int id, char * filename = NULL);
    virtual ~Shape3D () {};

    void setRadius (float r);
    void setHeight (float h);
    void setWireframe (bool wf);

    void draw (Matrix4& C);
    void draw_noCull (Matrix4& C) {draw (C);}
    void calcBdSphere ();
    float max (float num1, float num2);
    void updateBdSphere (Matrix4& transf);
    int intersectBdSphere (Vector3& normal, Vector3& pt);



  protected:

    // Store a ptr to a 3D obj
    Shape * _self;
    int _shape_id;

    float _radius, _height;
    bool _isWireframe;

    void init ();

    // center and radius of bounding sphere
    //   1st pair is wrt local coords.
    //   2nd pair is after transformation matrix is applied.
    Vector3 _bdsphere_c, _trans_bdsphere_c;
    float _bdsphere_r, _trans_bdsphere_r;

    ObjFile objFile;
    char _objFilename [PATH_MAX];

    float minmax [6];
};

static int obj_cellW = 5, obj_cellH = 5;


#endif

