// 
// (Mabel) Mengzi Zhang
// 30 Oct 2010
// CSE 167 hw 5
//

#include <transformGroup.h>
#include <shape3D.h>


// -------------------------------------------------------------- ctor dtor --

TransformGroup::TransformGroup ()
{
  // Group's
  size = 0;
  capacity = 1;
  childList = new Node *;

  // TransformGroup's
  mat = Matrix4 ();

  // Node's
  name = new char [name_maxlen];
  *name = 0;

  *className = 0;
  strcat (className, "TransformGroup");
}

TransformGroup::~TransformGroup ()
{
  if (childList)
    delete[] childList;
}


// ---------------------------------------------------------------- getters --

Matrix4 * TransformGroup::getMat ()
{
  return &mat;
}


// ---------------------------------------------------------------- setters --

void TransformGroup::setFrusNear (Vector3& n, Vector3& p)
{
  for (int i = 0; i < 3; i++)
  {
    norm_near.set (i, n [i]);
    pt_near.set (i, p [i]);
  }
}

void TransformGroup::setFrusFar (Vector3& n, Vector3& p)
{
  for (int i = 0; i < 3; i++)
  {
    norm_far.set (i, n [i]);
    pt_far.set (i, p [i]);
  }
}

void TransformGroup::setFrusLeft (Vector3& n, Vector3& p)
{
  for (int i = 0; i < 3; i++)
  {
    norm_left.set (i, n [i]);
    pt_left.set (i, p [i]);
  }
}

void TransformGroup::setFrusRight (Vector3& n, Vector3& p)
{
  for (int i = 0; i < 3; i++)
  {
    norm_right.set (i, n [i]);
    pt_right.set (i, p [i]);
  }
}

void TransformGroup::setFrusTop (Vector3& n, Vector3& p)
{
  for (int i = 0; i < 3; i++)
  {
    norm_top.set (i, n [i]);
    pt_top.set (i, p [i]);
  }
}

void TransformGroup::setFrusBott (Vector3& n, Vector3& p)
{
  for (int i = 0; i < 3; i++)
  {
    norm_bott.set (i, n [i]);
    pt_bott.set (i, p [i]);
  }
}

/*
void TransformGroup::setFrustumPlanes (Vector3 ** norms, Vector3 ** pts)
{
  fprintf (stderr, "in TG::setFrusPlanes()\n");

  if ((! norms) || (! pts))
  {
    fprintf (stderr, "Error: TransformGroup::setFrustumPlanes(): "
      "one or more params is NULL\n");
    return;
  }

  for (int i = 0; i < 6; i ++)
  {
    frus_norms[i] = new Vector3 ();
    frus_pts [i] = new Vector3 ();

    fprintf (stderr, "in loop\n");

    if (! norms+i)
      fprintf (stderr, "Error: TransformGroup::setFrustumPlanes(): "
        "norms[%d] is NULL\n", i);
    else if (! frus_norms+i)
      fprintf (stderr, "Error: TransformGroup::setFrustumPlanes(): "
        "frus_norms[%d] is NULL\n", i);
    else

      // setting the pointer in array
      frus_norms[i] = norms[i];
      frus_pts[i] = pts[i];

fprintf (stderr, "TG:setFrusPlanes(): param norm  %f %f %f \n",
  frus_norms[i]->get(0), frus_norms[i]->get(1), frus_norms[i]->get(2));

fprintf (stderr, "TG:setFrusPlanes(): norm %f %f %f pt %f %f %f \n",
  frus_norms[i]->get(0), frus_norms[i]->get(1), frus_norms[i]->get(2),
  frus_pts[i]->get(0), frus_pts[i]->get(1), frus_pts[i]->get(2)
  );

  }

}
*/


// -------------------------------------------------------------------- fns --

// Apply transformation matrix to the subtree below this node, i.e
//   all children inside childList [].
void TransformGroup::draw (Matrix4& C)
{
  // Matrix multiplication C_new = CM. New mat at this node is on the right,
  //   `.` when go up from obj, mats are attached to the left of eqn. So
  //   now when traversing down, attach them to the right.
  //   (M1(M2(M3p))) = (M1M2M3)p. Normally you think about it the LHS way,
  //   because we need to do fn calls, we use associative property and do it
  //   the RHS way.
  Matrix4 C_new = Matrix4 ();
  C.mul (mat, C_new);

  Node * curr_child;
  Shape3D * curr_shape;
  int count = 0;

  // TODO: To optimize, this only has to be done when camera moves (e.g.
  //   perspective changes, or frustum updated), if the culling status of each
  //   obj is kept from frame to frame. This saves some calculation when the
  //   perspective is static, and increases FPS when static!
  // Loop through all children in childList[], apply transformation matrix to
  //   each by calling draw() and passing in the matrix so far
  for (int i = 0; i < size; i ++)
  {
    curr_child = childList [i];

    // Debug. Very helpful, goes through entire hierarchy!
    //fprintf (stderr, "Calling %s.draw()\n", curr_child -> getName());


    // If child is type Shape3D, test if this child's bding sphere intersects
    //   frustum
    if  (strcmp (curr_child -> getClassName (), "Shape3D") == 0)
    {
//fprintf (stderr, "culling test starting\n");

      // Calculate local bding sphere and update it with the
      //   transformation matrix
      ((Shape3D *) curr_child) -> calcBdSphere ();
      ((Shape3D *) curr_child) -> updateBdSphere (C_new);

/*
      // For each of the 6 planes of the frustum, intersect bding sphere of 
      //   this obj with frustum plane.
      // If bding sphere is inside frustum or intersecting frustum, draw it.
      //   Else, bding sphere is entirely outside frustum, don't draw (cull).
      for (int i = 0; i < 6; i++)
      {
fprintf (stderr, "%dth plane\n",i);

//if (! frus_norms [0])
//  fprintf (stderr, "NULL\n");
//else
//  fprintf (stderr, "in TransformGroup::draw() %f %f %f \n",
//    frus_norms[i]->get(0), frus_norms[i]->get(1), frus_norms[i]->get(2));

        // Sanity check
        if (frus_norms [i])
          // Intersect plane w/ bding sphere, accumulate ret val
          count += ((Shape3D *) curr_child) -> intersectBdSphere (
            *(frus_norms [i]), *(frus_pts) [i]);
      }
*/

      int n = 0,f=0,t=0,b=0,l=0,r=0;

      n = ((Shape3D *) curr_child) -> intersectBdSphere (
        norm_near, pt_near);
      f = ((Shape3D *) curr_child) -> intersectBdSphere (
        norm_far, pt_far);
      t = ((Shape3D *) curr_child) -> intersectBdSphere (
        norm_top, pt_top);
      b = ((Shape3D *) curr_child) -> intersectBdSphere (
        norm_bott, pt_bott);
      l = ((Shape3D *) curr_child) -> intersectBdSphere (
        norm_left, pt_left);
      r = ((Shape3D *) curr_child) -> intersectBdSphere (
        norm_right, pt_right);

      count = n + f + t + b + l + r;


      //if (count < 6)
      //if (count == -6)
//        fprintf (stderr,
//          "near %d far %d top %d bottom %d left %d right %d count %d\n",
//          n, f, t, b, l, r, count);
      //fprintf (stderr, "count = %d\n", count);


      // count == 6: sphere is entirely inside frustum. render it.
      // count > -6 && count < 6: sphere intersects frustum. render it.
      // count == -6, sphere entirely outside frustum, don't draw.
      //if (count > -6)
      if (count >= 0)
        curr_child -> draw (C_new);

//fprintf (stderr, "culling test done\n");
    }

    // For other types of children, just draw anyway, to apply transform mat
    //   to the group
    else
    {
      curr_child -> draw (C_new);
    }

    // reset
    count = 0;

  }

}


// Draw all geometry, No culling involved
void TransformGroup::draw_noCull (Matrix4& C)
{
  Matrix4 C_new = Matrix4 ();
  C.mul (mat, C_new);

  for (int i = 0; i < size; i ++)
  {
    // Debug. Very helpful, goes through entire hierarchy!
    //fprintf (stderr, "Calling %s.draw()\n", curr_child -> getName());

    childList[i] -> draw_noCull (C_new);
  }
}


void TransformGroup::mulRotate (float x, float y, float z, float a)
{
  Matrix4 rot = Matrix4 ();
  rot.rotate (x, y, z, a);
  rot.mul (mat, mat);
}

void TransformGroup::mulTranslate (float x, float y, float z)
{
  Matrix4 trans = Matrix4 ();
  trans.translate (x,y,z);
  trans.mul (mat, mat);
}

