// 
// (Mabel) Mengzi Zhang
// 30 Oct 2010
// CSE 167 hw 5
//

#ifndef _SHAPE3D_H_
#define _SHAPE3D_H_

// scenegraph
#include <leaf.h>
#include <node.h>

#include <shape.h>
#include <Vector3.h>
#include <objReader.h>

#define SHAPE_SPHERE 1
#define SHAPE_CUBE 2
#define SHAPE_CONE 3
#define SHAPE_OBJ 4
#define SHAPE_QUAD 5
#define SHAPE_TREE 6

// From hw 2
class ObjFile : public Shape
{
  public:

    ObjFile ();

    float min (float num1, float num2);

    void loadObj (float * minmax, Matrix4& transformMat, char * objFilename);
    void normalize (float * minmax, Matrix4& transformMat);
    void draw ();

  protected:

    // Vertices
    // v
    int nVerts;
    float * verts;
    // vn
    float * normals;
    // vt
    float * texcoords;
    // f. Index of 3 vertices that make up a triangle
    int nIndices;
    int * indices;
};


class Shape3D : public Leaf
{
  public:
    // Use virtual keyword to make sure the right destructor is called, i.e.
    //   TransformGroup's dtor, not Group (parent)'s detor.
    Shape3D ();
    Shape3D (Shape * s);
    Shape3D (int id, char * filename = NULL);
    virtual ~Shape3D () {};

    void setRadius (float r);
    void setHeight (float h);
    void setWireframe (bool wf);
    void setBlend (bool b);
    void setStatic (bool s);

    void draw (Matrix4& C);
    void draw_noCull (Matrix4& C) {draw (C);}
    void draw_Quad(float width, float height);
    void draw_Bush();
    void draw_Tree(float x1, float y1,
                   float x2, float y2,
                   float angle,
                   int n);
    void calcBdSphere ();
    float max (float num1, float num2);
    void updateBdSphere (Matrix4& transf);
    int intersectBdSphere (Vector3& normal, Vector3& pt);

    // Because scene graph can't make call to doDraw() with a bunch of args,
    //   when using Shape in scenegraph, first set all these vars, then call
    //   world->draw(), then the draw in Shape() will call doDraw with these
    //   preset params correctly.
    void setTexDim (int w, int h) {_tex_w = w; _tex_h = h;}
    void set_sg_nVerts (int n) {_sg_nVerts = n;}
    void set_sg_verts (float * p) {_sg_verts = p;}
    void set_sg_nIndices (int n) {_sg_nIndices = n;}
    void set_sg_indices (int * p) {_sg_indices = p;}
    void set_sg_colors (float * p) {_sg_colors = p;}
    void set_sg_solidColor (float * p) {_sg_solidColor = p;}
    void set_sg_normals (float * p) {_sg_normals = p;}
    void set_sg_texCoords (float * p) {_sg_texCoords = p;}
    // Texture in scene graph
    void set_sg_textureOn (bool b) {_sg_textureOn = b;}
    void set_sg_texName (GLuint i) {_sg_texName = i;}
     // env mapping
    void set_sg_envMapOn (bool b) {_sg_envMapOn = b;}
    void set_face_files (char **f);
    void makeCubeMap();
    void loadFace(GLenum target, char *filename);
    void updateTexgen(void);
    void updateWrap(void);
    void set_mode(int);
    void set_wrap(int);
    void set_mipmaps(int);
    //display list
    void set_displayListOn(bool d);





  protected:

    // Store a ptr to a 3D obj
    Shape * _self;
    int _shape_id;

    float _radius, _height;
    bool _isWireframe;

    void init ();

    // center and radius of bounding sphere
    //   1st pair is wrt local coords.
    //   2nd pair is after transformation matrix is applied.
    Vector3 _bdsphere_c, _trans_bdsphere_c;
    float _bdsphere_r, _trans_bdsphere_r;

    ObjFile objFile;
    char _objFilename [PATH_MAX];

    float minmax [6];

    // To conform to scene graph Node class fns
    int _tex_w, _tex_h;
    int _sg_nVerts, _sg_nIndices;
    int * _sg_indices;
    float * _sg_verts, * _sg_colors, * _sg_solidColor, * _sg_normals,
      * _sg_texCoords;
    bool _sg_textureOn;
    bool _sg_envMapOn;
    GLuint _sg_texName;

    // For blending translucent objs
    bool _blendOn;
// For opengl display list
    bool _isStatic;
    bool _dispListCompiled;
    GLuint _dispListIdx;
    bool _dispListOnOff;
	
	
    // Face file for env mapping
    char *faceFile[6];
    //targets for binding
    GLuint thisTexture;
   //GL_REFLECTION_MAP_EXT or GL_NORMAL_MAP_EXT
   int mode;
   //GL_CLAMP or GL_REPEAT
   int wrap;
   int mipmaps;


    void drawCalls ();
};

static int obj_cellW = 5, obj_cellH = 5;
static GLenum faceTarget[6] = {
    GL_TEXTURE_CUBE_MAP_POSITIVE_X_EXT,
    GL_TEXTURE_CUBE_MAP_NEGATIVE_X_EXT,
    GL_TEXTURE_CUBE_MAP_POSITIVE_Y_EXT,
    GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_EXT,
    GL_TEXTURE_CUBE_MAP_POSITIVE_Z_EXT,
    GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_EXT
  };



#endif

