//
// Mengzi Zhang
// 25 Sept 2010
// CSE 167 hw 1
//

#include <Vector4.h>


// Default ctor
Vector4::Vector4 ()
{
  num_elts = 4;
  for (int i = 0; i < num_elts; i++)
    v[i] = 0;
}

// ctor
Vector4::Vector4 (double v0, double v1, double v2, double v3)
{
  num_elts = 4;
  v[0] = v0;
  v[1] = v1;
  v[2] = v2;
  v[3] = v3;
}


// ------------------------------------------------------- getters, setters --

void Vector4::set (int idx, double val)
{
  if ((idx < 0) || (idx >= num_elts))
  {
    fprintf (stderr,
      "Error: Vector4::set(): index %d out of bounds, setting to -1\n", idx);
    v[idx] = -1;
  }

  v[idx] = val;
}

void Vector4::setAll (double v1, double v2, double v3, double v4)
{
  v[0] = v1;
  v[1] = v2;
  v[2] = v3;
  v[3] = v4;
}

double Vector4::get (int idx)
{
  if ((idx < 0) || (idx >= num_elts))
  {
    fprintf (stderr,
      "Error: Vector4::get(): index %d out of bounds, returning -1\n", idx);
    return -1;
  }

  return v[idx];
}

// Overload get() with operator []
double Vector4::operator[] (int idx)
{
  return get (idx);
}


// -------------------------------------------------------------------- ops --

// res = this_vec + vec2
void Vector4::add (Vector4& vec2, Vector4& res)
{
  for (int i = 0; i < num_elts; i++)
    res.set (i, v[i] + vec2[i]);
}

// Potential bug: is this the right way to return res? Make the ret val by
//   reference? If use Vector4& as ret type, then would be returning a local
//   var, will be out of scope...
Vector4 Vector4::operator+ (Vector4& vec2)
{
  Vector4 res = Vector4 ();
  add (vec2, res);
  return res;
}

// res = this_vec - vec2
void Vector4::sub (Vector4& vec2, Vector4& res)
{
  for (int i = 0; i < num_elts; i++)
    res.set (i, v[i] - vec2[i]);
}

Vector4 Vector4::operator- (Vector4& vec2)
{
  Vector4 res = Vector4 ();
  sub (vec2, res);
  return res;
}

// v.transpose * u
double Vector4::mul (Vector4& u)
{
  return v[0]*u[0] + v[1]*u[1] + v[2]*u[2] + v[3]*u[3];
}

// Dehomongenize; divides every elt by the last elt, then remove the last elt
//   (now equals 1).
void Vector4::dehom (Vector4& res)
{
  // Store last elt (4th)
  double last_elt = v[3];

  for (int i = 0; i < num_elts; i++)
  {
    // Potential bug: is this the right way to dehomogenize? If last elt is 0,
    //   then do nothing? `.` if divide, will get infinity as result. Then the
    //   matrix would be messed up.
    if (last_elt != 0)
      res.set (i, v[i] / last_elt);
  }

  // Check for floating pt errs
  if (v[3] != 1.0)
    v[3] = 1.0;
}

void Vector4::print ()
{
  fprintf (stderr, "(%f, %f, %f, %f)\n", v[0], v[1], v[2], v[3]);
}



