//
// (Mabel) Mengzi Zhang
// 21 Oct 2010
// CSE 167 hw 4
//

#include <GL/glut.h>
#include <light.h>

#define GL_LIGHT0_IDX 0
#define GL_LIGHT1_IDX 1
#define GL_LIGHT2_IDX 2
#define GL_LIGHT3_IDX 3
#define GL_LIGHT4_IDX 4
#define GL_LIGHT5_IDX 5
#define GL_LIGHT6_IDX 6
#define GL_LIGHT7_IDX 7

Light::Light ()
{
  litNum = GL_LIGHT0;

  for (int i = 0; i < 8; i ++)
    isLitOn [i] = false;

  pos = Vector4 ();
  dir = Vector3 ();

  // Node's
  name = new char [name_maxlen];
  *name = 0;
  *className = 0;
  strcat (className, "Shape3D");

}

// ---------------------------------------------------------------- Getters --




// ---------------------------------------------------------------- Setters --

void Light::enable ()
{
  glEnable (GL_LIGHTING);
}

void Light::enable (GLenum litNum)
{
  glEnable (litNum);
  updateState (litNum, true);
}

void Light::disable (GLenum litNum)
{
  glDisable (litNum);
  updateState (litNum, false);
}

bool Light::toggle (GLenum litNum)
{
  // Find the corresponding array index (0-7) to the GLenum light number,
  //   call routine to update state stored in array and enable/disable the
  //   light.
  switch (litNum)
  {
    case GL_LIGHT0:
      return toggle_update (0, litNum);
      //break;
    case GL_LIGHT1:
      return toggle_update (1, litNum);
      //break;
    case GL_LIGHT2:
      return toggle_update (2, litNum);
      //break;
    case GL_LIGHT3:
      return toggle_update (3, litNum);
      //break;
    case GL_LIGHT4:
      return toggle_update (4, litNum);
      //break;
    case GL_LIGHT5:
      return toggle_update (5, litNum);
      //break;
    case GL_LIGHT6:
      return toggle_update (6, litNum);
      //break;
    case GL_LIGHT7:
      return toggle_update (7, litNum);
      //break;
    default:
      return false;
      //break;
  }
}

bool Light::toggle_update (int idx, GLenum litNum)
{
  bool curr = isLitOn [idx];

  // If currently on, turn it off
  if (curr)
    glDisable (litNum);
  // If currently off, turn it on
  else
    glEnable (litNum);

  // Update state in array
  isLitOn [idx] = ! curr;

  return (!curr);
}

void Light::updateState (GLenum litNum, bool state)
{
  // Store state of 1 light, on or off. Useful for toggling lights
  switch (litNum)
  {
    case GL_LIGHT0:
      isLitOn [0] = state;
      break;
    case GL_LIGHT1:
      isLitOn [1] = state;
      break;
    case GL_LIGHT2:
      isLitOn [2] = state;
      break;
    case GL_LIGHT3:
      isLitOn [3] = state;
      break;
    case GL_LIGHT4:
      isLitOn [4] = state;
      break;
    case GL_LIGHT5:
      isLitOn [5] = state;
      break;
    case GL_LIGHT6:
      isLitOn [6] = state;
      break;
    case GL_LIGHT7:
      isLitOn [7] = state;
      break;
    default:
      break;
  }
}

void Light::setLitNum (GLenum num)
{
  litNum = num;
}

// For all lights
void Light::setAmbient (GLfloat v1, GLfloat v2, GLfloat v3, GLfloat v4)
{
  GLfloat vec [] = {v1, v2, v3, v4};
  glLightfv (litNum, GL_AMBIENT, vec);
}

void Light::setDiffuse (GLfloat v1, GLfloat v2, GLfloat v3, GLfloat v4)
{
  GLfloat vec [] = {v1, v2, v3, v4};
  glLightfv (litNum, GL_DIFFUSE, vec);
}

void Light::setSpec (GLfloat v1, GLfloat v2, GLfloat v3, GLfloat v4)
{
  GLfloat vec [] = {v1, v2, v3, v4};
  glLightfv (litNum, GL_SPECULAR, vec);
}

// Make v4 0 if want directional light,
//   make it 1 if want point light.
void Light::setPos (GLfloat v1, GLfloat v2, GLfloat v3, GLfloat v4)
{
  pos.setAll (v1, v2, v3, v4);

  GLfloat vec [] = {v1, v2, v3, v4};
  glLightfv (litNum, GL_POSITION, vec);
}

// For spotlight
void Light::setSpotDir (GLfloat v1, GLfloat v2, GLfloat v3)
{
  dir.setAll (v1, v2, v3);

  GLfloat vec [] = {v1, v2, v3};
  glLightfv (litNum, GL_SPOT_DIRECTION, vec);
}

void Light::setSpotExp (GLfloat v)
{
  glLightf (litNum, GL_SPOT_EXPONENT, v);
}

void Light::setSpotCut (GLfloat v)
{
  glLightf (litNum, GL_SPOT_CUTOFF, v);
}

// For attenuation
// Setting attenuation uses formula intensity = 1 / c1*k^2 + c2*k + c3
//   because of the division and extra calcs, using attenuation may
//   decrease performance.
void Light::setAttenConst (GLfloat c3)
{
  glLightf (litNum, GL_CONSTANT_ATTENUATION, c3);
}

void Light::setAttenLin (GLfloat c2)
{
  glLightf (litNum, GL_LINEAR_ATTENUATION, c2);
}

void Light::setAttenQuad (GLfloat c1)
{
  glLightf (litNum, GL_QUADRATIC_ATTENUATION, c1);
}


// -------------------------------------------------------------------- fns --


void Light::draw (Matrix4& C)
{
  Vector4 newpos = Vector4 ();
  C.mul (pos, newpos);
  newpos.dehom (newpos);
  setPos (newpos[0], newpos[1], newpos[2], newpos[3]);
//  fprintf (stderr, "newpos %f %f %f %f", newpos[0], newpos[1], newpos[2],
//    newpos[3]);

/*
  Vector3 newdir = Vector3 ();
  C.mul (dir, newdir);
  newdir.norm (newdir);
  setSpotDir (newdir[0], newdir[1], newdir[2]);
  fprintf (stderr, "newdir %f %f %f", newdir[0], newdir[1], newdir[2]);
*/
}



