//
// (Mabel) Mengzi Zhang
// 24 Oct 2010
// CSE 167 hw 4
//
// Fragment shader program for spot light source.
// Effects applied on gl_LightSource [1], i.e. GL_LIGHT1.
// Takes varying params output from vertex program and compute attenuation.
// final color = ambientGlobal + attenuation * (ambient + diffuse + specular)
// spotEffect = (spotDirection * lightDir) ^ spotExp
//

// Fragment shader for per-pixel diffuse shading. The
// teapot is a good object to test this shader, since
// the .obj file includes surface normals.

// The shader computes the dot product between the unit
// surface normal and light direction, which were 
// passed as varying inputs from the vertex shader. The
// result is multiplied with the vertex color, which is 
// accessed through a pre-defined varying variable.

// Outputs from vertex program, inputs to this fragment prgm
varying vec4 diffuse, ambientGlobal, ambient;
varying vec3 normal, lightDir, halfVector;
varying float dist;

void main()
{
  // Note: A fragment shader can't write a varying variable, hence we need a
  //   new var to store the normalized interpolated normal

  // normal of this fragment to obj surface?, halfVector of light,
  //   viewVector, lightDirection
  vec3 n, halfV, viewV, ldir;
  // Dot product of normal & light direction, of normal and half vector of
  //   light
  float NdotL, NdotHV;
  // Init. Later will add to each attenuated color component at this px
  vec4 color = ambientGlobal;
  // Attenuation
  float att;

  // Normalize interpolated normal from vertex prgm
  n = normalize (normal);

  // Compute dot prod btwn normal and normalized lightDir
  // Q: what is it??
  NdotL = max (dot (n, normalize (lightDir)), 0.0);

  if (NdotL > 0.0)
  {
    // (spotLightDirection / vector to this px) should >
    //   cos (spotLightDir / cone edge vector r), in order for this px to be
    //   in the cone (drew on paper to see, verified).
    float spotEffect = dot (normalize (gl_LightSource [1].spotDirection),
                      normalize (-lightDir));

    // If the point is within the cone of the spotlight
    //   (else don't need to color at all, it'll only have the effect of the
    //   global ambient light, no local light source).
    if (spotEffect > gl_LightSource [1].spotCosCutoff)
    {
      // Illumination calculation all same except added spotEffect component
      //   in attenuation.

      spotEffect = pow (spotEffect, gl_LightSource [1].spotExponent);
      // attenuation = 1 / (k0 + k1d + k2d^2), where d is distance btwn
      //   light source and this px on obj
      att = spotEffect / (gl_LightSource [1].constantAttenuation +
                       gl_LightSource [1].linearAttenuation * dist +
                       gl_LightSource [1].quadraticAttenuation * dist * dist);

      // final color = ambientGlobal + att (ambient + diffuse + specular), her
      //   only doing ambient + diffuse, not specular yet.
      // diffuse is multed by NdotL because...? diffuse * normal dot light dir,
      //   what does that mean?
      color += att * (diffuse * NdotL + ambient);

      halfV = normalize (halfVector);
      // Take max, in case dot prod is neg, then use 0
      NdotHV = max (dot (n, halfV), 0.0);
      // Add the specular component. Raise specular of light to a power, where
      //   power = shininess.
      //   obj specular * light specular * (NdotHV ^ shininess)
      color += att * gl_FrontMaterial.specular * gl_LightSource [1].specular *
             pow (NdotHV, gl_FrontMaterial.shininess);
    }
  }

  // Final color output
  gl_FragColor = color;

  //gl_FragColor = gl_LightSource [1].diffuse *
  //  max (dot (normal, normalize (lightDir)), 0.0) *
  //  gl_FrontMaterial.diffuse;
}

