//
// (Mabel) Mengzi Zhang
// 26 November 2010
// Final Project
//
// L-System plants
//

#ifndef _LSYS_H_
#define _LSYS_H_

#include <Vector3.h>
#include <shape3D.h>

#define NUM_VARS 1
#define NUM_CONSTS 6
#define MAX_RPL_SIZE 80
// Assume that each variable in grammar is 1 char long. Used when generate
//   replacement strings
#define VAR_SIZE 1

#define TYPE_VAR 1
#define TYPE_CONST 2

#define ACTION_NONE 1
#define ACTION_DRAW 2
#define ACTION_TURN 3
#define ACTION_PUSH 4
#define ACTION_POP 5

// Stack for push/pop in grammar.
// Each incremental # elts when need to extend capacity of stack by realloc
#define STACK_INC 2

#define DRAW_LEN_DEFAULT 10

#define TURN_NONE 0
#define TURN_X 1
#define TURN_Y 2
#define TURN_Z 3
#define ANG_DEFAULT 90 //0

class LSys : public Leaf
{

  public:

    LSys ();
    ~LSys ();

    // overwrite parent fns
    void draw (Matrix4& C);
    void draw_noCull (Matrix4& C) {draw (C);}

    // setters 
    void setAxiom (char * axiom);
    void setVar (int idx, char v, int action);
    void setConst (int idx, char c, int action, float deg = 0.0f,
      int turn = TURN_Z);
    void setRpl (int idx, char * r);
    void setNumRecur (int n);
    void setDrawLen (float l);

    bool assertNumVars (int n);
    bool assertNumConsts (int n);

    // L Sys fns
    // Generate string for specified iteration, by spawning replacement rules
    void recurse (int num_iter);

  protected:

    // Member fields

    // grammar
    char _var [NUM_VARS];
    char _rpl [NUM_VARS] [MAX_RPL_SIZE];
    int _varAction [NUM_VARS];

    char _const [NUM_CONSTS];
    int _constAction [NUM_CONSTS];
    // 3 directions
    float _constTurnDeg [NUM_CONSTS];
    int _constTurnDir [NUM_CONSTS];

    // init starting string
    char _axiom [MAX_RPL_SIZE];

    // string in current iteration
    // Potential bug: would we have memory problems? It's a ptr
    char * _prevStr, * _currStr;

    float _currAng;
    int _currDir;

    // pts to form segment
    Vector3 _pt1, _pt2;

    // Length of each segment drawn
    float _drawLen;

    // Stack for push and pop in grammar
    // Arrays. Need to be on heap, so declare as ptrs
    Vector3 ** _stackPt;
    float * _stackDeg;
    int * _stackDir;
    int _stackSize, _stackCap;

    // Number of iterations of recursions to go
    int _numRecur;


    // Member functions

    // Parse final string, for each var or each const, call routine to process
    //   the var / const
    void parseString ();

    // Look up what this var needs to do, call fn to do it
    void processStr (int type, int idx);

    // Actual drawing fns
    void rotate (float ang, int turn);
    void calcEndpt ();
    void drawForward ();
};

#endif

