//
// (Mabel) Mengzi Zhang
// 28 Nov 2010
// CSE 167 Final Project
//
// Particle System
//

#ifndef _PARTSYS_H_
#define _PARTSYS_H_

#include "leaf.h"
#include "Matrix4.h"

#define NUM_PARTICLES 1000 //10000

class PartSys : public Leaf
{

  public:

    PartSys ();
    ~PartSys ();

    // scene graph fns
    void draw (Matrix4& C);
    void draw_noCull (Matrix4& C) {draw (C);}

    // user interaction
    void decreaseSpray ();
    void increaseSpray ();
    void decreaseNumParticles ();
    void increaseNumParticles ();

    void setBillboard (Vector3& v, float ang);
    void setYaw (float y);
    void setPitch (float p);

  protected:

    // Particle structure
    // Each particle has a unique position, dir vec, color, and life
    typedef struct
    {
      float xPos, yPos, zPos;
      // direction vector of movement
      float xVec, yVec, zVec;
      float r, g, b, life;
    } SpriteInfo;


    // Store all the particles
    SpriteInfo Spr [NUM_PARTICLES];

    int _maxSpread, _maxParticles;

    float Spread, Angle;

    // For time-based movement, so that speed is independent of GPU's FPS/speed
    //   on diff machines, ctrled by time instead
    // prevTick: tick count of last frame
    // currTick: tick count of current frame
    long prevTick, currTick;
    // # ticks elapased btwn frames (currTick - prevTick), used to calc mvment,
    //   will mult all movement factors by it
    long numTicks;
 


    //bool LoadTexture(char *TexName, GLuint TexHandle);

    // floor plane
    float _floor_xmin, _floor_xmax;
    float _floor_zmin, _floor_zmax;
    float _floor_y;

    // for billboarding particles to face user
    float _viewYaw, _viewPitch;


    // shape the particles into a fire shape
    float _radius;

    // texture
    // Handles to our textures
    GLuint Texture[128];

    // user movement matrix, used to calculate yaw pitch roll for billboard
    Vector3 _billbVec;
    float _billbAng;


    bool LoadTexture (char * TexName, GLuint TexHandle);
};

#endif

