//
// Mengzi Zhang
// 24 Sept 2010
// CSE 167 hw 1
//
// This file is changed from the one in hw1. set() and get() now do [col][row]
//   instead of [row][col] as before.
//

#include <Matrix4.h>
#include <BasicMath.h>

#define num_rows 4
#define num_cols 4

// default ctor
// Init as identity matrix
Matrix4::Matrix4()
{
  for (int i = 0; i < 16; i++)
    m[i] = 0;
  m[0] = 1;
  m[5] = 1;
  m[10] = 1;
  m[15] = 1;
}

// supplied fn
// ctor
Matrix4::Matrix4(
  double m0, double m1, double m2, double m3, double m4, double m5, double m6,
  double m7, double m8, double m9, double m10, double m11, double m12,
  double m13, double m14, double m15
  )
{
  m[0] = m0;
  m[1] = m1;
  m[2] = m2;
  m[3] = m3;
  m[4] = m4;
  m[5] = m5;
  m[6] = m6;
  m[7] = m7;
  m[8] = m8;
  m[9] = m9;
  m[10] = m10;
  m[11] = m11;
  m[12] = m12;
  m[13] = m13;
  m[14] = m14;
  m[15] = m15;
}

// supplied fn
// Get ptr to mat elts data
double* Matrix4::getPointer()
{
  return &m[0];
}

// supplied fn
// Ret a 4x4 identity matrix
void Matrix4::identity()
{
  for (int i = 0; i < 16; i++)
    m[i] = 0;
  m[0] = 1;
  m[5] = 1;
  m[10] = 1;
  m[15] = 1;
}

// supplied fn
// Make rotatn mat wrt y axis. Just a pure rotation mat to mult other mats to?
// angle in degrees
// [cos   0 sin  0
//  0     1 0    0
//  -sin  0 cos  0
//  0     0 0    1]
void Matrix4::rotateY(double angle)
{
  double th = BasicMath::radian(angle);
  m [0] = cos(th);
  m [1] = 0;
  m [2] = sin(th);
  m [3] = 0;
  m [4] = 0;
  m [5] = 1;
  m [6] = 0;
  m [7] = 0;
  m [8] = -sin(th);
  m [9] = 0;
  m [10] = cos(th);
  m [11] = 0;
  m [12]  = 0;
  m [13] = 0;
  m [14] = 0;
  m [15] = 1;
}

// Make rotation matrix about x axis
//   1 0    0
//   0 cos -sin 
//   0 sin  cos
void Matrix4::rotateX (double angle)
{
  double th = BasicMath::radian(angle);
  
  m [0] = 1;
  m [1] = 0;
  m [2] = 0;
  m [3] = 0;
  m [4] = 0;
  m [5] = cos (th);
  m [6] = -sin (th);
  m [7] = 0;
  m [8] = 0;
  m [9] = sin(th);
  m [10] = cos (th);
  m [11] = 0;
  m [12] = 0;
  m [13] = 0;
  m [14] = 0;
  m [15] = 1;
}

// Make rotation matrix about z axis
//   cos -sin  0
//   sin  cos  0
//   0    0    1
void Matrix4::rotateZ (double angle)
{
  double th = BasicMath::radian(angle);

  m [0] = cos(th);
  m [1] = sin(th);
  m [2] = 0;
  m [3] = 0;
  m [4] = -sin(th);
  m [5] = cos (th);
  m [6] = 0;
  m [7] = 0;
  m [8] = 0;
  m [9] = 0;
  m [10] = 1;
  m [11] = 0;
  m [12] = 0;
  m [13] = 0;
  m [14] = 0;
  m [15] = 1;
}

// Potential bug: is this the right way to do it? See Fall 09 lec1 slide p.40,
//   there's a mat with direct formula we can copy
//void Matrix4::rotate (Vector3& axis, double angle)
// Using params is better than Vector3, `.` don't need array axis.
void Matrix4::rotate (double ax, double ay, double az, double angle)
{
  // Normalize vec a = [ax, ay, az] to be a unit vector
  //fprintf (stderr, "Matrix4::rotate: pre %f %f %f\n", ax, ay, az);
  Vector3 a = Vector3 (ax, ay, az);
  a.norm (a);
  ax = a[0];
  ay = a[1];
  az = a[2];
  //fprintf (stderr, "Matrix4::rotate: post %f %f %f\n", ax, ay, az);


  double th = BasicMath::radian (angle);
  //fprintf (stderr, "a.len %f angle %f th %f\n", a.len (), angle, th);

  // This works.
  // Matrix from slides don't work, distorts shape. This one is from Wikipedia
  //   and it works.
  m [0] = cos(th) + ax*ax * (1-cos(th));
  m [1] = ay * ax * (1 - cos(th)) + az * sin(th);
  m [2] = az * ax * (1 - cos(th)) - ay * sin(th);
  m [3] = 0;
  m [4] = ax * ay * (1 - cos(th)) - az * sin(th);
  m [5] = cos(th) + ay*ay * (1 - cos(th));
  m [6] = az * ay * (1 - cos(th)) + ax * sin(th);
  m [7] = 0;
  m [8] = ax * az * (1 - cos(th)) + ay * sin(th);
  m [9] = ay * az * (1 - cos(th)) - ax * sin(th);
  m [10] = cos(th) + az*az * (1 - cos(th));
  m [11] = 0;
  m [12] = 0;
  m [13] = 0;
  m [14] = 0;
  m [15] = 1;
}

void Matrix4::set (int idx, double val)
{
  m[idx] = val;
}

void Matrix4::setAll (double v0, double v1, double v2, double v3,
  double v4, double v5, double v6, double v7, double v8, double v9,
  double v10, double v11, double v12, double v13, double v14, double v15)
{
  m [0] = v0;
  m [1] = v1;
  m [2] = v2;
  m [3] = v3;
  m [4] = v4;
  m [5] = v5;
  m [6] = v6;
  m [7] = v7;
  m [8] = v8;
  m [9] = v9;
  m [10] = v10;
  m [11] = v11;
  m [12] = v12;
  m [13] = v13;
  m [14] = v14;
  m [15] = v15;
}

double Matrix4::get (int idx)
{
  return m[idx];
}

// THIS matrix times another matrix
// Caller needs to pass in matrix res to hold result. (else var created in this
//   fn will be out of scope after return, .`. can't just create a var in this
//   fn and return it! Will be lost).
// Parameters:
//   mat2 is unchanged; need to call get() so can't make it const.
// Returns: result is returned in res.
void Matrix4::mul (Matrix4& mat2, Matrix4& res)
{
  
  double m0 = m[0];
  double m1 = m[1];
  double m2 = m[2];
  double m3 = m[3];
  double m4 = m[4];
  double m5 = m[5];
  double m6 = m[6];
  double m7 = m[7];
  double m8 = m[8];
  double m9 = m[9];
  double m10 = m[10];
  double m11 = m[11];
  double m12 = m[12];
  double m13 = m[13];
  double m14 = m[14];
  double m15 = m[15];

  double n0 = mat2.getPointer() [0];
  double n1 = mat2.getPointer() [1];
  double n2 = mat2.getPointer() [2];
  double n3 = mat2.getPointer() [3];
  double n4 = mat2.getPointer() [4];
  double n5 = mat2.getPointer() [5];
  double n6 = mat2.getPointer() [6];
  double n7 = mat2.getPointer() [7];
  double n8 = mat2.getPointer() [8];
  double n9 = mat2.getPointer() [9];
  double n10 = mat2.getPointer() [10];
  double n11 = mat2.getPointer() [11];
  double n12 = mat2.getPointer() [12];
  double n13 = mat2.getPointer() [13];
  double n14 = mat2.getPointer() [14];
  double n15 = mat2.getPointer() [15];
  
  // col 1
  res.set (0, m0*n0+m4*n1+m8*n2+m12*n3);
  res.set (1, m1*n0+m5*n1+m9*n2+m13*n3);
  res.set (2, m2*n0+m6*n1+m10*n2+m14*n3);
  res.set (3, m3*n0+m7*n1+m11*n2+m15*n3);

  res.set (4, m0*n4+m4*n5+m8*n6+m12*n7);
  res.set (5, m1*n4+m5*n5+m9*n6+m13*n7);
  res.set (6, m2*n4+m6*n5+m10*n6+m14*n7);
  res.set (7, m3*n4+m7*n5+m11*n6+m15*n7);
  
  res.set (8, m0*n8+m4*n9+m8*n10+m12*n11);
  res.set (9, m1*n8+m5*n9+m9*n10+m13*n11);
  res.set (10, m2*n8+m6*n9+m10*n10+m14*n11);
  res.set (11, m3*n8+m7*n9+m11*n10+m15*n11);

  res.set (12, m0*n12+m4*n13+m8*n14+m12*n15);
  res.set (13, m1*n12+m5*n13+m9*n14+m13*n15);
  res.set (14, m2*n12+m6*n13+m10*n14+m14*n15);
  res.set (15, m3*n12+m7*n13+m11*n14+m15*n15);
}


// THIS matrix times Vec4
void Matrix4::mul (Vector4& vec, Vector4& res)
{
  double v0 = vec[0], v1 = vec[1], v2 = vec[2], v3 = vec[3];

  // Potential bug: assuming it's column vector, multiplied by each row of
  //   matrix. If it's supposed to be a row vector, then it'd be different.
  //   It'd be multiplying the columns of a matrix... 0 1 2 3 instead of 0 4
  //   8 12, etc.
  res.set (0, m[0]*v0+m[4]*v1+m[8]*v2+m[12]*v3);
  res.set (1, m[1]*v0+m[5]*v1+m[9]*v2+m[13]*v3);
  res.set (2, m[2]*v0+m[6]*v1+m[10]*v2+m[14]*v3);
  res.set (3, m[3]*v0+m[7]*v1+m[11]*v2+m[15]*v3);
}

void Matrix4::premul (Vector4& vec, Vector4& res)
{
  double v0 = vec[0], v1 = vec[1], v2 = vec[2], v3 = vec[3];

  res.set (0, v0*m[0]+v1*m[1]+v2*m[2]+v3*m[3]);
  res.set (1, v0*m[4]+v1*m[5]+v2*m[6]+v3*m[7]);
  res.set (2, v0*m[8]+v1*m[9]+v2*m[10]+v3*m[11]);
  res.set (3, v0*m[12]+v1*m[13]+v2*m[14]+v3*m[15]);
}


// THIS matrix times Vec3
// Vec3 is promoted to a Vec4, multiplication performed, then normalized to
//   return a Vec3.
void Matrix4::mul (Vector3& vec_orig, Vector3& res_fin)
{
  // Promote Vector3 to a Vector4
  Vector4 vec = Vector4 (vec_orig[0], vec_orig[1], vec_orig[2], 1);
  Vector4 res = Vector4 ();

  // mult matrix by Vector4 promoted from orig Vector3
  mul (vec, res);

  // Dehomogenize (divide by len) and return a Vector3
  res.dehom (res);

  // Drop 4th elt (== 1 after dehomogenize), i.e. ignore the 4th row of the
  //   4x4 matrix.
  for (int i = 0; i < 3; i++)
    res_fin.set (i, res[i]);
}

void Matrix4::premul (Vector3& vec_orig, Vector3& res_fin)
{
  // Promote Vector3 to a Vector4
  Vector4 vec = Vector4 (vec_orig[0], vec_orig[1], vec_orig[2], 1);
  Vector4 res = Vector4 ();

  // mult matrix by Vector4 promoted from orig Vector3
  premul (vec, res);

  // Dehomogenize (divide by len) and return a Vector3
  res.dehom (res);

  // Drop 4th elt (== 1 after dehomogenize), i.e. ignore the 4th row of the
  //   4x4 matrix.
  for (int i = 0; i < 3; i++)
    res_fin.set (i, res[i]);
}


// Does multiplication without dehomogenizing. Used for hyperbolic
//   interpolation in rasterization.
// Return w.
double Matrix4::mul_nodehom (Vector3& vec_orig, Vector3& res_fin)
{
  Vector4 vec = Vector4 (vec_orig[0], vec_orig[1], vec_orig[2], 1);
  Vector4 res = Vector4 ();

  mul (vec, res);
  // res.dehom (res);

  for (int i = 0; i < 3; i++)
    res_fin.set (i, res[i]);

  // Return w value
  return res[3];
}

void Matrix4::translate (double tx, double ty, double tz)
{
  m [0] = 1;
  m [1] = 0;
  m [2] = 0;
  m [3] = 0;
  m [4] = 0;
  m [5] = 1;
  m [6] = 0;
  m [7] = 0;
  m [8] = 0;
  m [9] = 0;
  m [10] = 1;
  m [11] = 0;
  m [12] = tx;
  m [13] = ty;
  m [14] = tz;
  m [15] = 1;
}

void Matrix4::scale (double sx, double sy, double sz)
{
  m [0] = sx;
  m [1] = 0;
  m [2] = 0;
  m [3] = 0;
  m [4] = 0;
  m [5] = sy;
  m [6] = 0;
  m [7] = 0;
  m [8] = 0;
  m [9] = 0;
  m [10] = sz;
  m [11] = 0;
  m [12] = 0;
  m [13] = 0;
  m [14] = 0;
  m [15] = 1;
}


//
// 0 4 8  12
// 1 5 9  13
// 2 6 10 14
// 3 7 11 15
void Matrix4::transpose (Matrix4& res)
{
  double * ptr = getPointer ();
  double * res_ptr = res.getPointer ();

  double orig0 = ptr [0];
  double orig1 = ptr [1];
  double orig2 = ptr [2];
  double orig3 = ptr [3];
  double orig4 = ptr [4];
  double orig5 = ptr [5];
  double orig6 = ptr [6];
  double orig7 = ptr [7];
  double orig8 = ptr [8];
  double orig9 = ptr [9];
  double orig10 = ptr [10];
  double orig11 = ptr [11];
  double orig12 = ptr [12];
  double orig13 = ptr [13];
  double orig14 = ptr [14];
  double orig15 = ptr [15];

  res_ptr [0] = orig0;
  res_ptr [1] = orig4;
  res_ptr [2] = orig8;
  res_ptr [3] = orig12;
  res_ptr [4] = orig1;
  res_ptr [5] = orig5;
  res_ptr [6] = orig9;
  res_ptr [7] = orig13;
  res_ptr [8] = orig2;
  res_ptr [9] = orig6;
  res_ptr [10] = orig10;
  res_ptr [11] = orig14;
  res_ptr [12] = orig3;
  res_ptr [13] = orig7;
  res_ptr [14] = orig11;
  res_ptr [15] = orig15;
}

void Matrix4::print ()
{
  fprintf (stderr, "[");

  // Just do 0-16 elts. To see col-mat that OpenGL sees. Going downwards.
  for (int i = 0; i < num_rows * num_cols; i++)
    fprintf (stderr, "%f, ", m [i]);
  fprintf (stderr, "\n");
}


// Inverse of matrix
// Note: scaling is not implemented, only rotation and translation components
void Matrix4::inv (Matrix4& res) 
{

  // Rotation portion
  double m0 = m [0];
  double m1 = m [1];
  double m2 = m [2];
  double m3 = m [3];
  double m4 = m [4];
  double m5 = m [5];
  double m6 = m [6];
  double m7 = m [7];
  double m8 = m [8];
  double m9 = m [9];
  double m10 = m [10];
  double m11 = m [11];
  double m12 = m [12];
  double m13 = m [13];
  double m14 = m [14];
  double m15 = m [15];

  // Transpose 3x3 rotation portion (rotation mat's inverse is its transpose)
  //   of this matrix
  res.set (0, m0);
  res.set (1, m4);
  res.set (2, m8);
  res.set (4, m1);
  res.set (5, m5);
  res.set (6, m9);
  res.set (8, m2);
  res.set (9, m6);
  res.set (10, m10);


  // Translation portion

  // Negate translation vector portion (RHS column) of 4x4 mat
  Vector3 tran = Vector3 ();
  tran.set (0, -m12);
  tran.set (1, -m13);
  tran.set (2, -m14);

  double res0 = res.getPointer() [0];
  double res1 = res.getPointer() [1];
  double res2 = res.getPointer() [2];
  double res4 = res.getPointer() [4];
  double res5 = res.getPointer() [5];
  double res6 = res.getPointer() [6];
  double res8 = res.getPointer() [8];
  double res9 = res.getPointer() [9];
  double res10 = res.getPointer() [10];

  // Multiply translation vec by transposed rotation mat, i.e. combine both
  //   inverses to make final inverse.
  Vector3 v2 = Vector3 ();
  v2.set (0, res0 * tran[0] + res4 * tran[1] + res8 * tran[2]);
  v2.set (1, res1 * tran[0] + res5 * tran[1] + res9 * tran[2]);
  v2.set (2, res2 * tran[0] + res6 * tran[1] + res10 * tran[2]);

  // Put new translation vector into last col of new 4x4 matrix
  res.set (12, v2[0]);
  res.set (13, v2[1]);
  res.set (14, v2[2]);

  // Make last row like iden matrix so can use with OpenGL
  res.set (3, 0.0f);
  res.set (7, 0.0f);
  res.set (11, 0.0f);
  res.set (15, 1.0f);
}

