#ifndef _SHAPE_H_
#define _SHAPE_H_

// Local
#include <BasicMath.h>
#include <Matrix4.h>
#include <leaf.h>

// opengl
#include <GL/glut.h>

class Shape : public Leaf
{
  public:
    const static int PATH_MODE_CIRC = 0;
    const static int PATH_MODE_ELLP = 1;
    const static int SOL_CYL = 1;
    const static int SOL_CONE = 2;

    Shape();   // Constructor
    Matrix4& getMatrix();
    void setMatrix (Matrix4 mat);
    void setScale (double sx, double sy, double sz);
    void setRotate (double rx, double ry, double rz, double rw);
    void setMulRotate (double rx, double ry, double rz, double rw);

    void setDepth (int d);
    void setTexDim (int w, int h);
    void spin(double);      // spin cube [degrees]
    void spin (double ax, double ay, double az, double deg);
    void rotate (double rx, double ry, double rz, double ang);
    void scale (double sx, double sy, double sz);
    void trans (double tx, double ty, double tz);
    void mulRotate (double rx, double ry, double rz, double ang);
    void mulTrans (double tx, double ty, double tz);
    void mulScale (double sx, double sy, double sz);

    // Draw shapes from given array of vertices and indices of vertices that
    //   form a triangle, and array of colors
    void doDraw (int nVerts, float * verts, int nIndices,
      int * indices, float * colors, float * solidColor = NULL,
      float * normals = NULL, float * texCoords = NULL);

    //void drawCirc_old (double cx, double cz, double r, double b,
    //  int x_step, int mode);
    // Central function, draws round, cylinder, or cone
    void drawCirc (double cx, double cy, double cz, double r, double b,
      int x_step, int mode, int isCyl = 0, float solHeight = 0.0,
      bool doTexture = false);
    void drawCircVert (double cx, double cy, double cz, double r, double b,
      int x_step, int mode);

    void drawTri_tex (Vector3& v1, Vector3& v2, Vector3& v3,
      Vector3& t1, Vector3& t2, Vector3& t3);
    void drawTri (Vector3& v1, Vector3& v2, Vector3& v3);
    void drawTri_clr (Vector3& v1, Vector3& v2, Vector3& v3,
      Vector3& color1, Vector3& color2, Vector3& color3);
    void drawTri_clr_norm (Vector3& v1, Vector3& v2, Vector3& v3,
      Vector3& color1, Vector3& color2, Vector3& color3,
      Vector3& norm1, Vector3& norm2, Vector3& norm3);
    void drawQuad (Vector3& v1, Vector3& v2, Vector3& v3, Vector3& v4);
    void drawVert (Vector3& vert);
    void drawVert_clr_norm (Vector3& vert, Vector3& color, Vector3& norm);

    void updateMat ();

    // To conform to scene graph Node class fns
    void draw (Matrix4& C)
    {
      // Transform matrix
      glLoadIdentity ();
      glMultMatrixd (C.getPointer());

      // Texture
      if (_sg_textureOn)
      {
        glEnable (GL_TEXTURE_2D);
        glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
        glBindTexture (GL_TEXTURE_2D, _sg_texName);
      }
      else
        glDisable (GL_TEXTURE_2D);

      // Draw geometry
      doDraw (_sg_nVerts, _sg_verts, _sg_nIndices, _sg_indices, _sg_colors,
        _sg_solidColor, _sg_normals, _sg_texCoords);
    }
    void draw_noCull (Matrix4& C) {draw (C);}
    // Because scene graph can't make call to doDraw() with a bunch of args,
    //   when using Shape in scenegraph, first set all these vars, then call
    //   world->draw(), then the draw in Shape() will call doDraw with these
    //   preset params correctly.
    void set_sg_nVerts (int n) {_sg_nVerts = n;}
    void set_sg_verts (float * p) {_sg_verts = p;}
    void set_sg_nIndices (int n) {_sg_nIndices = n;}
    void set_sg_indices (int * p) {_sg_indices = p;}
    void set_sg_colors (float * p) {_sg_colors = p;}
    void set_sg_solidColor (float * p) {_sg_solidColor = p;}
    void set_sg_normals (float * p) {_sg_normals = p;}
    void set_sg_texCoords (float * p) {_sg_texCoords = p;}
    // Texture in scene graph
    void set_sg_textureOn (bool b) {_sg_textureOn = b;}
    void set_sg_texName (GLuint i) {_sg_texName = i;}


  protected:

    Matrix4 matrix;                 // model matrix

    Matrix4 _mat_rot;
    Matrix4 _mat_trans;
    Matrix4 _mat_scale;

    double angle;                   // rotation angle [degrees]
    double _scalex, _scaley, _scalez;	// scaling the pts when drawVert
    double _rotx, _roty, _rotz, _rotw;

    int num_faces;
    int _depth;

    // Texture dimensions
    int _tex_w, _tex_h;

    // To conform to scene graph Node class fns
    int _sg_nVerts, _sg_nIndices;
    int * _sg_indices;
    float * _sg_verts, * _sg_colors, * _sg_solidColor, * _sg_normals,
      * _sg_texCoords;
    bool _sg_textureOn;
    GLuint _sg_texName;

};

#endif

