//
// (Mabel) Mengzi Zhang
// 8 Nov 2010
// CSE 167 hw 6 curves and surfaces
//

#include <stdio.h>
#include <stdlib.h>
#include <iostream>

// OpenGL
#include <GL/glut.h>

// Local
#include <curve.h>
#include <Matrix4.h>
#include <material.h>


using namespace std;




// ------------------------------------------------------------------- ctor --

Curve::Curve ()
{
  _errExists = false;
}



// ---------------------------------------------------------------- getters --

vector <Vector3 *> * Curve::getVerts () {return &_verts;}
vector <int> * Curve::getTriIndices () {return &_triIndices;}
vector <Vector3 *> * Curve::getNormals () {return &_norms;}
vector <Vector3 *> * Curve::getTexCoords () {return &_texCoords;}
Vector3 * Curve::getCtrlpts () {return _ctrlpts;}


// -------------------------------------------------------------------- fns --

// Define a generatrix, a 2D curve, in xy plane
// Parameters:
//   num_curves: number of Bezier segments
void Curve::generate2DCurve (int num_curves, Vector3 * ctrlpts,
  int num_ctrlpts)
{

  _num_curves = num_curves;
  _num_ctrlpts = num_ctrlpts;

  if (_num_ctrlpts != (_num_curves - 1) * 3 + 4)
  {
    _errExists = true;
    fprintf (stderr, "Error: # (%d) of ctrl pts does not match the "
      "number required (%d) "
      "for %d cubic Bezier curves.\n", _num_ctrlpts,
      (_num_curves - 1) * 3 + 4, _num_curves);
    return;
  }
  else
  {
    _errExists = false;
    _ctrlpts = ctrlpts;
  }

}


// Given control pts of curve, eval _num_samplepts pts on the curve
//   Formula (calculation of x from param t and pts p0 p1 p2 p3) is from Curves
//   slides p.19
void Curve::evalSamplePts (int num_samplepts)
{
  if (_errExists)
  {
    fprintf (stderr, "A previous error exists. Cannot proceed until fixed.\n");
    return;
  }

  _num_samplepts = num_samplepts;

  // Bernstein coefficients used to linearly interpolate ctrl pts to get pt x
  //   on curve
  float b0 = 0.0, b1 = 0.0, b2 = 0.0, b3 = 0.0;

  // 4 current ctrl pts
  Vector3 p0, p1, p2, p3;

  Matrix4 Gbez = Matrix4 ();
  Matrix4 Bbez = Matrix4 (-1,3,-3,1, 3,-6,3,0, -3,3,0,0, 1,0,0,0);
  // C = Gbez * Bbez
  Matrix4 C = Matrix4 ();
  Vector4 T = Vector4 ();
  Vector4 X = Vector4 ();
  // For normals. Deriv
  Vector4 Tprime = Vector4 ();
  Vector4 Xprime = Vector4 ();

  // pt on curve
  Vector3 * x;
  // normal of pt
  Vector3 * xprime;

  // Parameter of eqn. Fraction where new pt should be btwn prev lvl's 2 pts
  //   goes from 0.0 to 1.0 with a step determined by num_samplepts
  float t = 0.0, t2 = 0.0, t3 = 0.0;

  float t_step = 1.0f / (float) _num_samplepts;

  // For each curve segment making up the bezier curve, calculate all sample
  //   pts
  for (int i = 0; i < _num_curves; i ++)
  {
    // 4 control pts for this curve segment
    // Use 3i, not 4i, so that last pt of prev curve overlaps with 1st pt of
    //   next pt
    p0 = _ctrlpts [3*i];
    p1 = _ctrlpts [3*i+1];
    p2 = _ctrlpts [3*i+2];
    p3 = _ctrlpts [3*i+3];

    // Each column is a pi, row1 x, row2 y, row3 z. (from slide p.31)
    Gbez.setAll (p0[0], p0[1], p0[2], 0, p1[0], p1[1], p1[2], 0, p2[0], p2[1],
      p2[2], 0, p3[0], p3[1], p3[2], 1);

    // C = Gbez * Bbez
    Gbez.mul (Bbez, C);

    // First pt on curve == 1st ctrl pt (t = 0.0)
    x = new Vector3 (p0 [0], p0 [1], p0 [2]);
    _verts.push_back (x);

    // Normal of first pt (t = 0.0)
    // T' = [3t^2 2t 1 0], when t = 0.0, T' = [0 0 1 0]
    Tprime.setAll (0.0, 0.0, 1.0, 0.0);
    C.mul (Tprime, Xprime);
    xprime = new Vector3 (-Xprime[1], Xprime[0], Xprime[2]);
    xprime -> norm (*xprime);
    _norms.push_back (xprime);

    // Set 1st t for this curve. 0 already done, do the next one
    t = t_step;


    // Generate a pt on curve for each t
    //   j goes from 1 to last-1, i.e. skip first and last pts, `.` they are on
    //   ctrl pts given, no need to calculate.
    for (int j = 1; j < _num_samplepts - 1; j ++)
    {
      // parameter squared and cubed
      t2 = t * t;
      t3 = t2 * t;

      // T = [t^3 t^2 t 1] vertical column
      T.setAll (t3, t2, t, 1);

      // x(t) = CT
      C.mul (T, X);

      // Do in matrix form now, don't need this. Keep it in case mat form
      //   doesn't work. This is slides p. 19
      /*
      // Bernstein coeffs
      b0 = -t3 + 3*t2 - 3*t + 1;
      b1 = 3*t3 - 6*t2 + 3*t;
      b2 = -3*t3 + 3*t2;
      b3 = t3;

      // x(t) = b0p0 + b1p1 + b2p2 + b3p3
      p0.mul (b0, p0);
      p1.mul (b1, p1);
      p2.mul (b2, p2);
      p3.mul (b3, p3);

      // Calculate pt x on curve, add to _verts array for keeps
      x = new Vector3 ();
      p0.add (p1, *x);
      x -> add (p2, *x);
      x -> add (p3, *x);
      */

      // Take 1st 3 coords. 4th should be 1 or 0 anyway, don't need to dehom
      x = new Vector3 (X [0], X [1], X [2]);
      _verts.push_back (x);


      // Calc normals

      // T' = [3t^2 2t 1 0] vertical column
      Tprime.setAll (3*t2, 2*t, 1, 0);

      // X' = CT'
      C.mul (Tprime, Xprime);

      // Take 1st 3 coords. 4th should be 0 anyway because Tprimt[3] == 0
      //   tangent is (x', y', 0), normal is then (-y', x', 0). So flip x y.
      xprime = new Vector3 (-Xprime[1], Xprime[0], Xprime[2]);
      xprime -> norm (*xprime);
      _norms.push_back (xprime);


      // For next round
      t += t_step;
    }

    // Last pt on curve == last ctrl pt (t = 1.0)
    x = new Vector3 (p3 [0], p3 [1], p3 [2]);
    _verts.push_back (x);

    // Normal of last pt on curve. (t = 1.0)
    // T' = [3t^2 2t 1 0], when t = 1.0, T' = [3 2 1 0]
    Tprime.setAll (3.0, 2.0, 1.0, 0.0);
    C.mul (Tprime, Xprime);
    xprime = new Vector3 (-Xprime[1], Xprime[0], Xprime[2]);
    xprime -> norm (*xprime);
    _norms.push_back (xprime);
  }

  // Debug
  fprintf (stderr, "DEBUG: Final t (should be very close to 1.0): %f\n", t);
  fprintf (stderr, "DEBUG: Number of pts in _verts (should == %d): %d\n",
    _num_curves *_num_samplepts, _verts.size ());
  fprintf (stderr, "DEBUG: Number of normals: %d\n", _norms.size());
  //for (int i = 0; i < 10; i++)
  //  _verts.at(i) -> print ();



  Vector3 * texcoord;
  float u = 0.0;
  _u_step = 1.0f / (_num_samplepts * _num_curves);

  // 1st pt in col, do manually, 0.0
  texcoord = new Vector3 (0.0, 0.0, 0.0);
  _texCoords.push_back (texcoord);

  // Add texture coords to all current (1st col) pts. Do from 1 to 2nd-to-last
  //   pt, `.` 0th (u=0.0) and last (u=1.0) pt do manually to avoid floating
  //   pt errs.
  for (int i = 1; i < _verts.size () - 1; i++)
  {
    u += _u_step;
    texcoord = new Vector3 (u, 0.0, 0.0);
    _texCoords.push_back (texcoord);
  }

  // last pt in col, do manually, 1.0
  texcoord = new Vector3 (1.0, 0.0, 0.0);
  _texCoords.push_back (texcoord);

  fprintf (stderr, "DEBUG: Number of texcoords: %d\n", _texCoords.size());

}


// Parameters:
//   ang_inc: angle increment for a full 360 degree rotation around y-axis
void Curve::rotateAroundY (float ang_inc, bool fillCaps)
{
  if (_errExists)
  {
    fprintf (stderr, "A previous error exists. Cannot proceed until fixed.\n");
    return;
  }

  _angle_inc = ang_inc;

  // Rotate all sample pts on curve around y-axis to generate multiple curves,
  //   store pts on each curve into _verts vector

  Vector3 * newpt, * newpt_prime;

  // Last rotation don't need to do, should just connect back to starting pts?
  int num_cols = (int) floor (360.0 / ang_inc);
  int num_pts_per_col = _num_samplepts * _num_curves;

  // Make the center vert that connects the top and bottom caps
  // top/bott: (y is just 1st sample pt's y. x is 0, z = 0.)
  // bott/top: (y is just LAST sample pt's y
  Vector3 * cent_cap_first = new Vector3 (0.0, _verts.at (0)->get(1), 0.0);
  Vector3 * cent_cap_last = new Vector3 (0.0,
    _verts.at (num_pts_per_col - 1)->get(1), 0.0);

  Matrix4 transf = Matrix4 ();

  int col1start = 0;
  int col1end = num_pts_per_col - 1;
  int col2start = col1end + 1;
  int col2end = col1end + num_pts_per_col;

  Vector3 * texcoord;
  float u = 0.0f, v = 0.0f;


  // Number of columns to spin.
  // Each iteration rotates sample pts for an additional ang_inc degrees
  //   start i with 1, `.` when rotate by i * ang_inc, pts rotated by 0 (when
  //   i=0) degrees have already been added to the array when calced sample pts
  // Get each pt from the first group in _verts, which should be the orig
  //   sample pts calculated from ctrl pts.
  // Generate triangle tesselation and store triangle indices into _triIndices
  //   vector
  for (int i = 1; i < num_cols + 1; i ++)
  {
    // Debug
    //fprintf (stderr, "col1start %d col1end %d col2start %d col2end %d\n",
    //  col1start, col1end, col2start, col2end);

    // reset
    u = 0.0f;

    // Each sample pt on the entire bezier curve (# segments * # sample pts per
    //   segment), rotate ang_inc to get a new pt
    for (int j = 0; j < num_pts_per_col; j ++)
    {
      // Generate new pts

      newpt = new Vector3 ();
      newpt_prime = new Vector3 ();

      // For last round, close gap by connecting back to very 1st column.
      //   Force this instead of rotation `.` to make sure there's no tiny gap
      //   due to floating pt errors
      if (i == num_cols)
      {
        // new pt of last round == very 1st col
        newpt -> setAll (_verts.at(j) -> get(0), _verts.at(j) -> get(1),
          _verts.at(j) -> get (2));

        // new normal of last round == very 1st col
        newpt_prime -> setAll (_norms.at (j) -> get(0),
          _norms.at (j) -> get(1), _norms.at (j) -> get(2));

        // For texcoords. Last round just rotates by 360 degrees, 360/360 = 1
        v = 1.0;
      }

      else
      {
        // Make a matrix rotated around y axis for ang_inc degrees
        transf.rotate (0.0, 1.0, 0.0, i * ang_inc);

        // Rotate curr pt around y axis
        transf.mul (*(_verts.at (j)), *newpt);

        // For normal. Rotate normal around y axis
        transf.mul (*(_norms.at (j)), *newpt_prime);
        // normalize so that normals are indeed normals
        newpt_prime -> norm (*newpt_prime);

        // For texcoords
        v = i * ang_inc / 360.0;
      }

      // Save pt
      _verts.push_back (newpt);
      _norms.push_back (newpt_prime);


      // Calculate texture coordinates

      texcoord = new Vector3 (u, v, 0.0);
      _texCoords.push_back (texcoord);

      // u = curve parameter t
      u += _u_step;

 
      // Add tesselation of triangles to indices array

      // Make the top cap
      if (j == 0)
      {
        if (fillCaps)
        {
          // Assumption:
          //   CENTER of cap at same height of 1st sample pt will be added after
          //   the loop to end of array. Cap goes from center, right bottom
          //   vert, left bottom vert.
          // add another num_pts_per_col to account for the last col that
          //   overlaps the 1st col (needed, to close the gap).
          _triIndices.push_back (num_pts_per_col * num_cols + num_pts_per_col);
          _triIndices.push_back (col2start);
          _triIndices.push_back (col1start);
        }
      }

      // After getting a pt in new column, tesselate triangle and save triangle
      //   indices into array
      // j goes from 1 to _num_samplepts - 1, i.e. 2nd dot to last dot, 1st dot
      //   j == 0 skip, `.` don't have a triangle yet.
      else if (j > 0)
      {
        // |\
        // |_\   CCW. p0 lower left, p1 upper left, p2 lower right
        _triIndices.push_back (col1start + j);
        _triIndices.push_back (col1start + j - 1);
        _triIndices.push_back (col2start + j);

        // __
        // \ |
        //  \|  CCW. p0 lower right, p1 upper left, p2 upper right 
        _triIndices.push_back (col2start + j);
        _triIndices.push_back (col1start + j - 1);
        _triIndices.push_back (col2start + j - 1);


        if (fillCaps)
        {
          // Make the bottom cap
          if (j == num_pts_per_col - 1)
          {
            // Assumption:
            //   CENTER of cap at same height of last sample pt will be added
            //   after the loop to end of array. Cap goes from center, right
            //   bottom vert, left bottom vert.
            // add another num_pts_per_col to account for the last col that
            //   overlaps the 1st col (needed, to close the gap).
            _triIndices.push_back (num_pts_per_col * num_cols + num_pts_per_col
              + 1);
            _triIndices.push_back (col1end);
            _triIndices.push_back (col2end);
          }
        }

      }
    }

    // Set for next round
    col1start = col2start;
    col1end = col2end;
    col2start = col2end + 1;
    col2end += num_pts_per_col;

    // Debug
    if (col1end != _verts.size () - 1)
      fprintf (stderr, "Error: Triangle tesslation: col1end %d should be "
        "equal to (current size - 1) %d of array of all pts in solid.\n",
        col1end, _verts.size ()-1);

  }

  // Put in center pts of the caps at the end of array
  if (fillCaps)
  {
    _verts.push_back (cent_cap_first);
    _verts.push_back (cent_cap_last);
  }


  // Debug
  fprintf (stderr, "Total number of triangles in solid: (should be int) %f\n",
    _triIndices.size () / 3.0);

  fprintf (stderr, "_verts.size() = %d\n", _verts.size());
  fprintf (stderr, "_norms.size() = %d\n", _norms.size());
  fprintf (stderr, "_texCoords.size() = %d\n", _texCoords.size());
}

// Go through all ctrl pts, compare tangents at each curve's endpoint with
//   tangent of next curve's start pt, if all pairs equal, return true.
//   else return false.
bool Curve::testC1Cont ()
{
  // Go through start point of each individual bezier curve, test for C1
  //   continuity at that start pt, by passing in the ctrl pts before it and
  //   after it.
  // Skip 0th ctrl pt, `.` there's no notion of continuous `.` there's no pts
  //   BEFORE it.
  // Skip last ctrl pt, similar reason, there's no pts AFTER it.
  for (int i = 3; i < _num_ctrlpts - 3; i+=3)
  {
    fprintf (stderr, "Ctrl pt [%d] ", i);

    // If any one pair is not C1 continuous, then this whole concatenated
    //   curve is not C1 continuous.
    if (! testC1Cont_single (*(_ctrlpts+i-1), *(_ctrlpts+i), *(_ctrlpts+i+1)) )
    {
      // Recheck this segment in next iteration
      i -= 3;
      fprintf (stderr, "Retesting this segment...");
    }
  }

  // If haven't returned false at this pt yet, then all bezier segments are
  //   C1 cont.
  fprintf (stderr, "Curve: all Bezier segments in this curve are C1 "
    "continuous.\n");
  return true;
}

// Point Pi is C1 continuous if Pi - P(i-1) == P(i+1) - Pi
bool Curve::testC1Cont_single (Vector3& p_prev, Vector3& p, Vector3& p_next)
{
  Vector3 res1 = Vector3 (), res2 = Vector3 ();
  
  // P(i) - P(i-1) => res1
  p.sub (p_prev, res1);

  // P(i+1) - P(i) => res2
  p_next.sub (p, res2);

  if (res1.equals (res2))
  {
    fprintf (stderr, "is C1 continuous.\n");
    return true;
  }
  else
  {
    fprintf (stderr, "is NOT C1 continuous. Adjusting...");

    // prev segment gets priority. Adjust 2nd segment to make res2 == res1.
    //   i.e. Adjust P(i+1) in next segment.
    //   want res2, currently = P(i+1) - P(i), to = res1
    //   => P(i+1) - P(i) = res1
    //   => P(i+1)        = res1 + P(i)
    //   So just set P(i+1) to res1 + P(i). Then should have c1 cont.
    res1.add (p, p_next);

    return false;
  }
}


