//
// (Mabel) Mengzi Zhang
// 26 November 2010
// Final Project
//
// L-System plants
//
// Grammar assumption:
//   Assume there is always a draw command after each turn of angle.
//   If that's not the case, e.g. you want to do 2 angle turns in a row before
//   a turn, such as things like > - F, where > and - are different turn angs,
//   and F is a draw,
//   then need to add code to rotate() and setEndpt() to allow that.
//   That's too much math calc for me right now, so I'm not doing it.
//
// Note1: Always allocate (using new [] on heap) 1 MORE than # chars for all 
//   char arrays, to store the '\0' char!! Else will get run-time error at
//   some point when delete[] calls free ().
// Note2: Always use delete[] for anything you allocate using new[]. Use
//   delete for anything you allocate using new (note without the array []).
//
//
/* --------------------------------------
Grammar 1:
(from 2nd midterm 2008)

Just some pattern

Axiom: B

Variables (NUM_VARS = 1):
B - draw a segment or execute a replacement rule

Replacement rule:
B <- A 45Z A [-45Z B][45Z B]

Constants (NUM_CONSTS = 5):
+ - rotate around Z-axis by # degrees (specify # when call setConst)
- - rotate around Z-axis by -# degrees (specify # when call setConst)
[ - push current position and angle onto stack
] - pop stack
A - draw a segment


----------------------------------------
Grammar 2:
(from Wikipedia L System fractal plant)

Fractal plant

Axiom:
X

Variables (NUM_VARS = 2):
X - no drawing action, only used for generating replacement grammar
F - draw forward

Constants (NUM_CONSTS = 4):
+ - turn right 25 deg
- - turn left 25 deg
[ - push current position and angle onto stack
] - pop stack

Replacement Rules:
X <- F-[[X]+X]+F[+FX]-X)
F <- FF

Angle: 25 deg


------------------------------------- */


#include <stdlib.h>
#include <stdio.h>
#include <iostream>

#include <GL/glut.h>

#include <lsys.h>


// ------------------------------------------------------------ ctor, dtor --

LSys::LSys ()
{
  _prevStr = NULL;
  _currStr = NULL;

  _pt1 = Vector3 ();
  _pt2 = Vector3 ();

  _drawLen = DRAW_LEN_DEFAULT;

  _currAng = ANG_DEFAULT;
  _currDir = TURN_NONE;

  for (int i = 0; i < NUM_VARS; i++)
  {
    _varAction [i] = -1;
  }

  for (int i = 0; i < NUM_CONSTS; i++)
  {
    _constAction [i] = -1;
    _constTurnDeg [i] = 0.0f;
    _constTurnDir [i] = TURN_Z;
  }

  // stack for push and pop
  _stackPt = NULL;
  _stackDeg = NULL;
  _stackDir = NULL;
  _stackSize = 0;
  _stackCap = 0;

  // Number of recursions to go
  _numRecur = 0;
}

LSys::~LSys ()
{
  if (_prevStr)
    delete[] _prevStr;
  if (_currStr)
    delete[] _currStr;
  if (_stackPt)
  {
    // Delete all Vector3s dynamically allocated
    for (int i = 0; i < _stackSize; i ++)
      delete (_stackPt [i]);

    delete[] _stackPt;
  }
  if (_stackDeg)
    delete[] _stackDeg;
  if (_stackDir)
    delete[] _stackDir;
}


// ---------------------------------------------------- overwrite hier fns --

void LSys::draw (Matrix4& C)
{

  glLoadIdentity ();
  glMultMatrixd (C.getPointer ());

  // draw the segments
  //recurse (2);
  recurse (_numRecur);

}




// --------------------------------------------------------------- setters --

// Parameters:
//   axiom: string that starts 0th iteration
//   axiomSize: # chars in axiom string
void LSys::setAxiom (char * axiom)
{
  *_axiom = 0;

  if (axiom)
  {
    //_axiomSize = strlen (_axiom);
    strcpy (_axiom, axiom);
  }
  else
  {
    //_axiomSize = 0;
    fprintf (stderr, "Warning: LSys::LSys(): axiom is NULL. Axiom is empty "
      "string, axiomSize set to 0\n");
  }
}


// Parameters:
//   draw: whether this var triggers a drawForward
void LSys::setVar (int idx, char v, int action)
{
  if (idx >= NUM_VARS)
  {
    fprintf (stderr, "Error: LSys::setConst: idx %d exceeds NUM_CONSTS %d, "
      "specifying too many constants.\n",
      idx, NUM_CONSTS);
    return;
  }

  _var [idx] = v;
  _varAction [idx] = action;
}

void LSys::setConst (int idx, char c, int action, float deg, int turn)
{
  if (idx >= NUM_CONSTS)
  {
    fprintf (stderr, "Error: LSys::setConst: idx %d exceeds NUM_CONSTS %d, "
      "specifying too many constants.\n",
      idx, NUM_CONSTS);
    return;
  }

  _const [idx] = c;
  _constAction [idx] = action;
  _constTurnDeg [idx] = deg;
  _constTurnDir [idx] = turn;
}

// Set replacement rule
// Parameters:
//   idx: which var is the replacement for
//   r: replacement string
// Note: size of string should not exceed MAX_RPL_SIZE
void LSys::setRpl (int idx, char * r)
{
  if (strlen (r) > MAX_RPL_SIZE)
  {
    fprintf (stderr, "Error: LSys::setRpl(): size of replacement string "
      "should not exceed MAX_RPL_SIZE %d\n", MAX_RPL_SIZE);
    return;
  }
  else if (r)
  {
    // store replacement string
    *_rpl [idx] = 0;
    strcat (_rpl [idx], r);
  }
}


void LSys::setDrawLen (float l)
{
  _drawLen = l;
}


void LSys::setNumRecur (int n)
{
  _numRecur = n;
}

bool LSys::assertNumVars (int n)
{
  if (n == NUM_VARS)
  {
    fprintf (stderr, "Asserted number of constants (%d).\n", n);
    return true;
  }
  else
  {
    fprintf (stderr, "Error: LSys::assertNumConsts(): "
      "number of constants assumed (%d) and NUM_VARS (%d) do NOT match. "
      "Solve it by changing NUM_VARS in src/lsys.h to %d\n",
      n, NUM_VARS, n);
    return false;
  }
}

bool LSys::assertNumConsts (int n)
{
  if (n == NUM_CONSTS)
  {
    fprintf (stderr, "Asserted number of constants (%d).\n", n);
    return true;
  }
  else
  {
    fprintf (stderr, "Error: LSys::assertNumConsts(): "
      "number of constants assumed (%d) and NUM_CONSTS (%d) do NOT match. "
      "Solve it by changing NUM_CONSTS in src/lsys.h to %d\n",
      n, NUM_CONSTS, n);
    return false;
  }
}


// ------------------------------------------------------------- L Sys fns --


// Generate new strings for this iteration by using the replacement rules
// Parameters:
//   num_iter: # of iterations to draw
void LSys::recurse (int num_iter)
{
  // To process tokens
  char * tokenPtr = NULL;
  char lastChar = '0';
  int currVarIdx = -1;

  // To replace strings
  // Need to be on heap. so just declare a ptr.
  char * currRpl = NULL;
  // tmpChar holds only 1 char, but needs 2 chars, to hold the '\0' char!!
  //   Else get run-time error when delete calls free()!
  char tmpChar [2];
  *tmpChar = 0;
  char * tmpStr = NULL;

  // init prevStr to be axiom
  if (_prevStr) delete[] _prevStr;
  _prevStr = new char [strlen (_axiom) + 1];
  strcpy (_prevStr, _axiom);

  // Make sure iteration 0 works
  if (_currStr) delete[] _currStr;
  _currStr = new char [strlen (_axiom) + 1];
  strcpy (_currStr, _axiom);

  // Reset
  _currAng = ANG_DEFAULT;
  _currDir = TURN_NONE;
  _pt1.setAll (0.0, 0.0, 0.0);
  _pt2.setAll (0.0, 0.0, 0.0);

  // Replace strings in each iteration, until generate final string
  for (int i = 0; i < num_iter; i ++)
  {
    // Clear currStr. Prev round's should already been saved in _prevStr.
    // Put on heap, `.` when don't know size of a var at programming time, put
    //   it on heap.
    if (_currStr)
      delete[] _currStr;
    _currStr = NULL;

    // Parse through previous string
    for (int j = 0; j < strlen (_prevStr); j ++)
    {
      // See if this char matches a variable
      currVarIdx = -1;
      for (int k = 0; k < NUM_VARS; k ++)
      {
        if (_prevStr [j] == _var [k])
        {
          currVarIdx = k;
          break;
        }
      }

      if (tmpStr)
      {
        delete[] tmpStr;
        tmpStr = NULL;
      }

      // If current char does NOT match a var, must be a const or space. Copy
      //   it directly to the new string.
      if (currVarIdx < 0)
      {
        // Concat 1 char, need to convert it to string first
        sprintf (tmpChar, "%c", _prevStr [j]);

        // If curr string empty, just allocate 2 chars
        // memory management for string. destroy and create a new string with
        //   length+1. Then copy orig content and curr char to new string.
        if (! _currStr)
        {
          tmpStr = new char [strlen (tmpChar) + 1];
          *tmpStr = 0;
        }

        // If curr string not empty, concat it first
        else
        {
          tmpStr = new char [strlen (_currStr) + strlen (tmpChar) + 1];
          *tmpStr = 0;
          strcat (tmpStr, _currStr);
        }
 
        // concat new char to string
        strcat (tmpStr, tmpChar);
      }

      // If current char matched a var, replace it with var's replacement
      else
      {
        // Construct new string
        if (! _currStr)
        {
          tmpStr = new char [strlen (_rpl [currVarIdx]) + 1];
          *tmpStr = 0;
        }
        else
        {
          // New string size = orig _currStr size (replacements gen-ed for
          //   prev tokens) + 1 (for the space) + new token's replacement's
          //   size
          tmpStr = new char [strlen (_currStr) + strlen (
            _rpl [currVarIdx]) + 1];
          *tmpStr = 0;
          strcat (tmpStr, _currStr);
          //strcat (tmpStr, " ");
        }

        // new replacement for this char
        strcat (tmpStr, _rpl [currVarIdx]);
      }

      // Save new str to _currStr
      if (_currStr)
      {
        delete[] _currStr;
        _currStr = NULL;
      }
      if (tmpStr)
      {
        _currStr = new char [strlen (tmpStr) + 1];
        strcpy (_currStr, tmpStr);
        delete[] tmpStr;
        tmpStr = NULL;
      }

      // Debug. Very useful. Prints out every step of new string generation
      //fprintf (stderr, "DEBUG: replacing chars %s\n", _currStr);

    }

    // Prepare for next iter. Save currStr as preStr for next round
    if (_prevStr) delete[] _prevStr;
    _prevStr = new char [strlen (_currStr) + 1];
    strcpy (_prevStr, _currStr);

    // Debug. Very useful. Shows the string in each iteration
    //fprintf (stderr, "DEBUG: iteration %d, _currStr = %s\n", i, _currStr);

  }

  // Draw final string into geometry
  parseString ();

}


// Parse final string and call routines to process vars / consts in string
void LSys::parseString ()
{
  // Debug. Very useful. Shows the final string that goes to draw geometry
  //fprintf (stderr, "DEBUG: LSys::parseString(): _currStr = %s\n", _currStr);

  int currVarIdx = -1;

  // work on this if want to use numbers to specify degrees to turn. now just
  //   using symbols, easier and fewer fn calls.
  //char * currStartPos = NULL;

  // Define a delimiter to skip when parsing, to increase speed
  const char DELIM_CHAR = ' ';

  // Parse through pattern string.
  // For each action variable / constant we see, do the action.
  for (int i = 0; i < strlen (_currStr); i++)
  {
    // If this char is just a delimiter, skip to next char
    if (_currStr [i] == DELIM_CHAR)
    {
      // pt to next char to process
      //currStartPos = _currStr + i + 1;
      continue;
    }

    // Try to match a variable
    currVarIdx = -1;
    for (int j = 0; j < NUM_VARS; j++)
    {
      // If this char matches a variable
      if (_currStr [i] == _var [j])
      {
        // Mark which variable this char is
        currVarIdx = j;
        processStr (TYPE_VAR, currVarIdx);
        break;
      }
    }

    // If this char matched a var, then done with this char
    //   (`.` processStr() should have been called already).
    if (currVarIdx >= 0)
    {
      // pt to next char to process
      //currStartPos = _currStr + i + 1;
      continue;
    }

    // If this char didn't match a var, see if it's a const
    for (int j = 0; j < NUM_CONSTS; j++)
    {
      if (_currStr [i] == _const [j])
      {
        // Mark which constant this char is
        currVarIdx = j;
        processStr (TYPE_CONST, currVarIdx);
        break;
      }
    }

    // pt to next char to process
    //currStartPos = _currStr + i + 1;

  }
}


// Parameters:
//   type: var or const
void LSys::processStr (int type, int idx)
{

  if ( ((type == TYPE_VAR) && (_varAction [idx] == ACTION_NONE)) ||
       ((type == TYPE_CONST) && (_constAction [idx] == ACTION_NONE)) )
    return;

  // If this const's action is to draw
  else if ( ((type == TYPE_VAR) && (_varAction [idx] == ACTION_DRAW)) ||
       ((type == TYPE_CONST) && (_constAction [idx] == ACTION_DRAW)) )
  {
    // Note:
    // Shouldn't set pts here. need to set it at end of draw(), then if hit
    //   a pop, then it gets reset. don't set it here, will overwrite whatever
    //   a pop just wrote.

    // Set default _pt2, for same DIRECTION as before (note that 0 angle does
    //   not necessarily go horizontal, it'll go along whatever dir the curr
    //   segment is pointing towards. So need to call calcEndpt() to do it,
    //   using the old _currAng.
    calcEndpt (); 

    // Draw using _pt1 and _pt2 set before hitting this variable / const
    drawForward ();
  }

  // If this const's action is to turn
  else if ( ((type == TYPE_CONST) && (_constAction [idx] == ACTION_TURN)) ||
            ((type == TYPE_VAR) && (_varAction [idx] == ACTION_TURN)) )
  {
    // Call routine to set _currAng
    rotate (_constTurnDeg [idx], _constTurnDir [idx]);
  }

  // If this const's action is to push endpoint onto stack
  else if ( ((type == TYPE_VAR) && (_varAction [idx] == ACTION_PUSH)) ||
            ((type == TYPE_CONST) && (_constAction [idx] == ACTION_PUSH)) )
  {
    // Data to save onto stack
    Vector3 * newpt = new Vector3 (_pt2 [0], _pt2 [1], _pt2 [2]);
    float newang = _currAng;
    int newdir = _currDir;

    // If stack over capacity, then need to reallocate and copy data
    if (_stackSize >= _stackCap)
    {
      // Update capacity for new stack
      _stackCap += STACK_INC;

      Vector3 ** tmpStackPt = new Vector3 * [_stackCap];
      float * tmpStackDeg = new float [_stackCap];
      int * tmpStackDir = new int [_stackCap];

      // Copy old content of stack to new stack
      for (int i = 0; i < _stackSize; i++)
      {
        tmpStackPt [i] = _stackPt [i];
        tmpStackDeg [i] = _stackDeg [i];
        tmpStackDir [i] = _stackDir [i];
      }

      // Move new tmp stack back to real member variable
      if (_stackPt) delete[] _stackPt;
      _stackPt = tmpStackPt;
      if (_stackDeg) delete[] _stackDeg;
      _stackDeg = tmpStackDeg;
      if (_stackDir) delete[] _stackDir;
      _stackDir = tmpStackDir;
    }

    // Add new pt to end of new stack
    _stackPt [_stackSize] = newpt;
    _stackDeg [_stackSize] = newang;
    _stackDir [_stackSize] = newdir;

    // Update size, just added 1 elt
    _stackSize ++;

  }

  else if ( ((type == TYPE_CONST) && (_constAction [idx] == ACTION_POP)) ||
            ((type == TYPE_VAR) && (_varAction [idx] == ACTION_POP)) )
  {
    // Save top of stack (tail of array) as startpt for drawing (_pt1)
    Vector3 poppedPt = *(_stackPt [_stackSize - 1]);
    _pt1.setAll (poppedPt [0], poppedPt [1], poppedPt [2]);

    // set pt2 to popped pt too, in case there's an immediate push (`.` push
    //   saves pt2, not pt1. So if there's a push immedaitely after a pop,
    //   in which case it should save the popped pt, which is only saved to
    //   pt1 so far, instead PUSH would save pt2 onto the stack. then the
    //   popped pt is lost. so we need to save the popped pt to both pt1 and
    //   pt2, so when there's immediate push, PUSH saves pt2, which would
    //   hold the popped pt.
    _pt2.setAll (poppedPt [0], poppedPt [1], poppedPt [2]);

    // save top of stack
    _currAng = _stackDeg [_stackSize - 1];
    _currDir = _stackDir [_stackSize - 1];
    // should only happen for the initial turn const
    //if (_currDir == TURN_NONE)
    //  _currDir = TURN_Z;

    // Clear array elt
    _stackPt [_stackSize - 1] = NULL;
    _stackDeg [_stackSize - 1] = 0.0f;
    _stackDir [_stackSize - 1] = TURN_NONE;

    // Update size
    _stackSize --;
  }

}


// Set _currAng (in degrees)
void LSys::rotate (float ang, int turn)
{
  // Accumulate the angle
  _currAng += ang;

  // If set to turn somewhere, set the var (else keep the current dir)
  if (turn != TURN_NONE)
    _currDir = turn;

  // chop off if extra round
  if (_currAng > 360)
    _currAng -= 360;
  else if (_currAng < -360)
    _currAng += 360;
}

// Set _pt2 according to current rotation degree
void LSys::calcEndpt ()
{
  // If rotate around Z, then it's on xy-plane, x = cos, y = sin, z doesn't
  //   change
  // Also put TURN_NONE here, `.` in beginning, before any turn has been set
  //   yet, the initial turn is TURN_NONE, and assume that's on xy-plane, so
  //   put it wrt to Z.
  if ((_currDir == TURN_Z) || (_currDir == TURN_NONE))
    _pt2.setAll (
      _pt1 [0] + _drawLen * cos (BasicMath::radian (_currAng)),
      _pt1 [1] + _drawLen * sin (BasicMath::radian (_currAng)),
      _pt1 [2]);

  // If rotate around Z, then it's on xy-plane, x = cos, z = sin, y doesn't
  //   change
  else if (_currDir == TURN_Y)
    _pt2.setAll (
      _pt1 [0] + _drawLen * cos (BasicMath::radian (_currAng)),
      _pt1 [1],
      _pt1 [2] + _drawLen * sin (BasicMath::radian (_currAng)));

  // If rotate around Z, then it's on xy-plane, z = cos, y = sin, x doesn't
  //   change
  else if (_currDir == TURN_X)
    _pt2.setAll (
      _pt1 [0],
      _pt1 [1] + _drawLen * sin (BasicMath::radian (_currAng)),
      _pt1 [2] + _drawLen * cos (BasicMath::radian (_currAng)));
}


// Draw a segment from _pt1 to _pt2
// Note: Set _pt1 and _pt2 before calling this fn
void LSys::drawForward ()
{
  glBegin (GL_LINES);
    //glColor3f(0, 255, 0);

    glVertex3f (_pt1 [0], _pt1 [1], _pt1 [2]);
    glVertex3f (_pt2 [0], _pt2 [1], _pt2 [2]);
  glEnd ();

  // Next startpt = prev endpt
  _pt1.setAll (_pt2 [0], _pt2 [1], _pt2 [2]);
}



