//
// (Mabel) Mengzi Zhang
// 28 Nov 2010
// CSE 167 Final Project
// Particle System
// 
// Modified from orig code from
//   http://gpwiki.org/index.php/OpenGL_Tutorial_Framework:Particles
//
// Things to control:
//   Length of shooting up, color of fire, size of particle
//
/*
YouTube demo's tips

Use blending functions to blend particles together:

Hi, what blendfunc did you use? I've had limited luck with GL_ONE, GL_ONE,
  since that makes the background shine through. How does your system act
  if there is a background (such as a checker pattern) ?
>I use GL_SRC_ALPHA, GL_ONE. using GL_ONE for the your particle will cause
  its peripheral area to be black and covering the particles below it. you
  want it to be transparent (alpha -> 0) instead. of black.

depending on what do you. choose the color of particles?
> the color I usually use is red = 1.0, green = 0.6, blue = 0.3. but it needs
  some fine tuning to achieve best results. add. more green and you'll get
  larger area of yellow. add more blue and you'll get larger area of white.

so... how did you do it? Is the. source available?
> sorry but no... it's. built upon a relatively large library that I don't
  completely own... but it's not complicated though, just a bunch of
  particles blended together

*/
//
//


#include <cmath>
#include <ctime>
#include <cstdlib>

#include <GL/glut.h>
 
//#include "Framework.h"
#include "tga.h"
#include "partsys.h"
#include "BasicMath.h"

// A few parameter definitions, tweak these to alter appearance of spray
const float PARTICLE_SIZE = 6.0f;//6.0f;//2.0f;//0.5f;
const int INITIAL_PARTICLE_SPREAD = (int) ceil (PARTICLE_SIZE) * 5000;
const float SPEED_DECAY = 0.00005f; // (Gravity) rate of fall/rise
// just pick a number to multiply/divide by, *5.0f, /0.05f, etc
const float HEIGHT_CONST = (float) (PARTICLE_SIZE / 0.05f);
 

// ------------------------------------------------------------- ctor, dtor --

PartSys::PartSys ()
{
  _floor_xmin = -10.0f;
  _floor_xmax = 10.0f;
  _floor_zmin = -10.0f;
  _floor_zmax = 10.0f;
  _floor_y = 0.0f;

  _viewYaw = 0.0f;
  _viewPitch = 0.0f;

  prevTick = currTick = clock ();
  //prevTick = 0;
  //currTick = 0;
  numTicks = 0;

  _maxSpread = INITIAL_PARTICLE_SPREAD;
  _maxParticles = NUM_PARTICLES / 2;

  Spread = 0.0f;
  Angle = 0.0f;

  // Init all particles to dead
  for (int i = 0; i < NUM_PARTICLES; i ++)
  {
    // Set life to 0, to indicate that each particle will be init by main loop
    Spr [i].life = 0.0f;

    Spr [i].xPos = 0.0f;
    Spr [i].yPos = 0.0f;
    Spr [i].zPos = 0.0f;

    Spr [i].xVec = 0.0f;
    Spr [i].yVec = 0.0f;
    Spr [i].zVec = 0.0f;

    // Set color vals, these will stay CONST throughout prgm
    Spr [i].r = 1.0f;
    Spr [i].g = 0.0f;
    Spr [i].b = 0.0f;
  }

  _radius = 3.0f * PARTICLE_SIZE;
 
  _billbVec = Vector3 ();
  _billbAng = 0.0f;

  // Seed the randomiser
  srand (time (NULL));

  // texture

  // Allocate all textures in one go
  glGenTextures (128, Texture);
 
  // Load the textures
  LoadTexture("in/particle.tga", Texture[0]);
  //LoadTexture("in/Marble.tga", Texture[1]);
 

}

PartSys::~PartSys ()
{
  // orig code does this after main loop ends, but since we have while(true)
  //   for that, when user presses F1, probably won't get to end of loop.
  // So put this in destructor, hopefully it still deletes.
  glDeleteTextures (128, Texture);
}



// ---------------------------------------------------------------- setters --

void PartSys::setYaw (float y) {_viewYaw = y;}
void PartSys::setPitch (float p) {_viewPitch = p;}

void PartSys::setBillboard (Vector3& v, float ang)
{
  //m.mul (_billbMat, _billbMat);

  /*
  _billbMat.setAll (m.get(0), m.get(1), m.get(2), m.get(3), m.get(4),
    m.get(5), m.get(6), m.get(7), m.get(8),
    m.get(9), m.get(10), m.get(11), m.get(12),
    m.get(13), m.get(14), m.get(15));
  */

  _billbVec.setAll (v[0], v[1], v[2]);
  _billbAng = ang;

}


// -------------------------------------------------------------------- fns --

// Here we go!
// Render particles
void PartSys::draw (Matrix4& C)
{
  // Handles to our textures
  //GLuint Texture[128];

  // Allocate all textures in one go
  //glGenTextures (128, Texture);
 
  // Setup our screen
  //glMatrixMode(GL_PROJECTION);
  //glViewport(0,0,800,600);
  //glLoadIdentity();
  //glFrustum(-.5f,.5f,-.5f*(600.0f/800.0f),.5f*(600.0f/800.0f),1.0f,500.0f);
  //glMatrixMode(GL_MODELVIEW);
 
  // Enable blending, to give the blurry effect of particles, instead of
  //   discrete quads or triangles geometry
  //glEnable (GL_BLEND);
  //glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  //glBlendFunc (GL_SRC_ALPHA, GL_ONE);
 
  // Load the textures
  //LoadTexture("../particle.tga",Texture[0]);
  //LoadTexture("../Marble.tga",Texture[1]);
  glEnable(GL_TEXTURE_2D);
 
  // Reset view
  glLoadIdentity(); 
  // for scenegraph hierarchy
  glMultMatrixd (C.getPointer ());
 
  // Get ticks since last frame
  currTick = clock ();
  numTicks = currTick - prevTick;
  //fprintf (stderr, "prevTick %ld currTick %ld numTicks %ld\n",
  //  prevTick, currTick, numTicks);
  //fprintf (stderr, "Spread %f Angle %f\n", Spread, Angle); 
  prevTick = currTick;

  // we have this in run.cpp's trackball. don't need it here.
  // Calculate view angles
  //_viewPitch = (Mouse.My-300)/4.0f;
  //_viewYaw = (Mouse.Mx-300)/3.0f;

  // we have trackball in run.cpp
  // Set up the view based on the mouse position
  //glTranslatef (0.0f, -10.0f, -50.0f);
  //glRotatef (_viewPitch, 1.0f, 0.0f, 0.0f);
  //glRotatef (_viewYaw, 0.0f, 1.0f, 0.0f);
 
  /* we don't need a floor plane
  // Set up for floor plane
  //glBindTexture(GL_TEXTURE_2D,Texture[1]);
  glColor4f(1.0f,1.0f,1.0f,1.0f);
  // Disable blend for floor plane
  //glDisable(GL_BLEND);

  // Draw floor plane
  glBegin(GL_QUADS);
    //glTexCoord2f(0.0f,1.0f);
    glVertex3f(-10.0f, 0.0f, 10.0f);
    //glTexCoord2f(1.0f,1.0f);
    glVertex3f( 10.0f, 0.0f, 10.0f);
    //glTexCoord2f(1.0f,0.0f);
    glVertex3f( 10.0f, 0.0f,-10.0f);
    //glTexCoord2f(0.0f,0.0f);
    glVertex3f(-10.0f, 0.0f,-10.0f);
  glEnd();
  */
 
 
  // Loop through each particle, if it's alive, update it; if it's dead,
  //   generate a new one.
  for(int i = 0; i < _maxParticles; i++)
  {
    // If particle is still alive, update pos, vec, and life
    if (Spr[i].life > 0.0f)
    {
      // update position according to direction vectors and ticks
      //Spr[i].xPos += (Spr[i].xVec * numTicks);
      //Spr[i].yPos += (Spr[i].yVec * numTicks);
      //Spr[i].zPos += (Spr[i].zVec * numTicks);
      Spr[i].xPos += (Spr[i].xVec);
      Spr[i].yPos += (Spr[i].yVec);
      Spr[i].zPos += (Spr[i].zVec);

      //Spr[i].yVec -= (SPEED_DECAY * numTicks);  
      // gravity pull. particles go down
      //Spr[i].yVec -= (SPEED_DECAY);
      // particles rise / vaporize
      Spr[i].yVec += (SPEED_DECAY);
 
      /*
      // 'Bounce' particle if on the floor square
      if (Spr[i].xPos > _floor_xmin && Spr[i].xPos < _floor_zmin &&
          Spr[i].zPos > _floor_zmin && Spr[i].zPos < _floor_zmax)
      {
        // If particle is intersecting floor plane
        if (Spr[i].yPos < _floor_y + PARTICLE_SIZE)
        {
          // Move particle up
          Spr[i].yPos = _floor_y + PARTICLE_SIZE;
          // Decrease life by fixed amount
          Spr[i].life -= 0.01f;
          // Reverse direction of movement, must have been neg before, make
          //   it positive big num, so it bounces up. Then next iteration,
          //   it'll drop again. Until its life runs out.
          Spr[i].yVec *= -0.6f;
        }
      }
      */

      // Change yellowness according to life
      //Spr[i].g = Spr[i].life;// * (0.2f + ((float) (rand() % 50) / 100.0f));
      Spr[i].g = Spr[i].life;//(0.2f + ((float) (rand() % 50) / 100.0f));

      // Decrease life wrt ticks
      //Spr[i].life -= (0.0001f*numTicks);
      // smaller constant means life is decrementing slower, larger height.
      //   bigger const means life decrements faster, so smaller height.
      Spr[i].life -= (0.00001f*numTicks);
    }

    // If particle is already dead, spawn a new particle
    else
    {
        // Reset position to origin (0,y,0). Start y from exactly on floor.
        //Spr[i].xPos = 0.0f;
        //Spr[i].yPos = _floor_y + PARTICLE_SIZE;
        //Spr[i].zPos = 0.0f;
        // Reset position to a round bottom of a sphere
        float theta = BasicMath::radian (
          (-180.0 / _maxParticles) * (float) i);
          //(-180.0 / _maxParticles) * (float) (
          //(rand() % ((_maxParticles + 1) * 10)) / 10.0f ));
        // this is exactly on the circle outline of semisphere
        //Spr[i].xPos = (float) _radius * cos (theta);
        //Spr[i].yPos = (float) _radius * sin (theta);
        // this is somewhere random on the semisphere
        Spr[i].xPos = (float) (rand() % (int) (_radius * 10.0f)) / 10.0f *
          cos (theta);// - (_radius / 2.0f);
        Spr[i].yPos = (float) (rand() % (int) (_radius * 10.0f)) / 10.0f *
          sin (theta);// - (_radius / 2.0f);
        // multiplying 10 gives you 10 times more random numbers to choose
        //   from. Then divide by 10 to get back to the range we want.
        Spr[i].zPos = (float) ((rand() % (int) (_radius * 10.0f)) / 10.0f) -
          (_radius / 2.0f);

        // Get a random spread and direction
        //Spread = (float) (rand() % _maxSpread) / 10000.0f;
        // dividing by 10000 is too tiny. need bigger.
        Spread = (float) (rand() % _maxSpread) / 1000.0f;
        // Random ang in quarter circle, 1.57 radians
        Angle = (float) (rand() % 157) / 100.0f;

        // Calculate X and Z vectors. spread is spherical radius, r cos gives
        //   horiz magnitude, r sin gives vert magnitude
        //Spr[i].xVec = cos (Angle) * Spread;
        //Spr[i].zVec = sin (Angle) * Spread;
        Spr[i].xVec = 0.0f;
        Spr[i].zVec = 0.0f;
 
        // Randomly reverse X and Z vector to neg, to complete the circle
        if (rand () % 2)
          Spr[i].xVec = - Spr[i].xVec;
        if (rand () % 2)
          Spr[i].zVec = - Spr[i].zVec;
 
        // Get a random initial falling speed
        // too small for large size particles
        //Spr[i].yVec = (float) (rand() % 500) / 10000.0f;
        Spr[i].yVec = (float) (rand() % 500) / HEIGHT_CONST;
 
        // Get a random life. it'll be btwn 0 and 1
        Spr[i].life = (float) (rand() % 100) / 100.0f;

        // Get a random yellowness
        Spr[i].g = 0.2f + ((float) (rand() % 50) / 100.0f);
    }
  }

  // Select particle texture
  glBindTexture (GL_TEXTURE_2D, Texture[0]);
  glEnable (GL_BLEND);
  //glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE);
  glDepthMask (GL_FALSE);

  // Particle size
  float partSize = 0.0f;

  // Loop thru & draw each particle
  for (int i = 0; i < _maxParticles; i++)
  {
    glPushMatrix();

    // Use life as alpha for blending. Shorter life = more faded
    glColor4f (Spr [i].r, Spr [i].g, Spr[i].b, Spr[i].life);

    //fprintf (stderr, "%f, %f, %f\n",
    //  Spr[i].xPos, Spr[i].yPos, Spr[i].zPos);
    //fprintf (stderr, "  %f, %f, %f\n",
    //  Spr[i].xVec, Spr[i].yVec, Spr[i].zVec);

    // Place the quad
    glTranslatef (Spr[i].xPos, Spr[i].yPos, Spr[i].zPos);

    // TODO: use trackball vals to do this... which means have to call a
    //   fn in here in displayCB()
    // Rotate particle to face viewer (billboard)
    glRotatef (-_viewYaw, 0.0f, 1.0f, 0.0f);
    //glRotatef (-_viewPitch, 1.0f, 0.0f, 0.0f);
    //glMultMatrixd (_billbMat.getPointer ());
    glRotatef (_billbAng, _billbVec[0], _billbVec[1], _billbVec[2]);
//fprintf (stderr, "%f %f %f %f\n", -_billbAng, _billbVec[0], _billbVec[1],
//  _billbVec[2]);

    // Assumption: life is always btwn 0 and 1. If it ever gets to negative
    //   then check the life-=numTicks part, it needs to multiply by the right
    //   factor to make numTicks btwn 0 and 1.
    if (Spr[i].life > 0.0f)
      partSize = PARTICLE_SIZE * Spr[i].life;
    else
      partSize = PARTICLE_SIZE;
    //fprintf (stderr, "%f\n", Spr[i].life);

    // Draw a quad to represent each particle
    glBegin (GL_QUADS);
      glTexCoord2f (0.0f, 0.0f);
      glVertex3f (-partSize, partSize, 0.0f);
      glTexCoord2f (0.0f, 1.0f);
      glVertex3f (-partSize, -partSize, 0.0f);
      glTexCoord2f (1.0f, 1.0f);
      glVertex3f (partSize, -partSize, 0.0f);
      glTexCoord2f(1.0f, 0.0f);
      glVertex3f (partSize, partSize, 0.0f);
    glEnd();

    glPopMatrix();
  }
   
  glDepthMask(GL_TRUE);
  glDisable (GL_BLEND);
  glDisable (GL_TEXTURE_2D);

  // Clean up textures
  //glDeleteTextures (128, Texture);
}

 
// Load a TGA texture
bool PartSys::LoadTexture (char * TexName, GLuint TexHandle)
{
  // Image loader
  TGAImg Img;
 
  // Load our Texture
  if (Img.Load (TexName) != IMG_OK)
    return false;
 
  // Set our Tex handle as current
  glBindTexture (GL_TEXTURE_2D, TexHandle);
 
  // Create the texture
  if (Img.GetBPP () == 24)
    glTexImage2D (GL_TEXTURE_2D, 0, 3, Img.GetWidth(), Img.GetHeight(),0,
      GL_RGB, GL_UNSIGNED_BYTE, Img.GetImg());
  else if (Img.GetBPP () == 32)
    glTexImage2D (GL_TEXTURE_2D, 0, 4, Img.GetWidth(), Img.GetHeight(),0,
      GL_RGBA, GL_UNSIGNED_BYTE, Img.GetImg());
  else
    return false;
 
  // Specify filtering and edge actions
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
 
  return true;
}



// --------------------------------------------------- external interaction --

// Tightens (decrease range of) spray
void PartSys::decreaseSpray ()
{
  _maxSpread--;
  if(_maxSpread<1)
    _maxSpread=1;
}

// Widens (increases range of) spray
void PartSys::increaseSpray ()
{
  _maxSpread++;
  if(_maxSpread>500)
    _maxSpread=500;
}

// Reduce particle count (takes a while to be noticeable)
void PartSys::decreaseNumParticles ()
{
  _maxParticles-=1;
  if(_maxParticles<1)
    _maxParticles=1;
}

void PartSys::increaseNumParticles ()
{
  _maxParticles+=1;
  if(_maxParticles>NUM_PARTICLES)
    _maxParticles=NUM_PARTICLES;
    
  // Kill new particle to ensure good start point
  Spr[_maxParticles].life=0.0f; 
}
 

