//
// (Mabel) Mengzi Zhang
// 12 Nov 2010
// CSE 167 hw 6 Bezier patch
//


#include <patch.h>



// Parameter:
//Patch::Patch (Curve * c1, Curve * c2)
Patch::Patch ()
{
  //curve1 = c1;
  //curve2 = c2;
  _errExists = false;
}



// ---------------------------------------------------------------- getters --

vector <Vector3 *> * Patch::getVerts () {return &_verts;}
vector <int> * Patch::getTriIndices () {return &_triIndices;}
vector <Vector3 *> * Patch::getNormals () {return &_norms;}
vector <Vector3 *> * Patch::getTexCoords () {return &_texCoords;}
Vector3 * Patch::getCtrlpts () {return _ctrlpts;}



// -------------------------------------------------------------------- fns --

// Parameters:
//   ctrlpts: an array of 16 ctrl pts
void Patch::generateSurface (Vector3 * ctrlpts, int num_ctrlpts)
{
  if (num_ctrlpts != 16)
  {
    _errExists = true;
    fprintf (stderr, "Error: # (%d) of ctrl pts does not match the "
      "number required (16)", num_ctrlpts);
    return;
  }
  else
    _ctrlpts = ctrlpts;
}

void Patch::buildSurface (int num_samplepts)
{
  _num_samplepts = num_samplepts;


  //vector <Vector3 *> * ctrlpts1 = curve1 -> getCtrlpts ();
  //vector <Vector3 *> * ctrlpts2 = curve2 -> getCtrlpts ();


  // B
  Matrix4 Bbez = Matrix4 (-1,3,-3,1, 3,-6,3,0, -3,3,0,0, 1,0,0,0);


  // G

  // Ctrl pts
  Vector3 p0 = _ctrlpts [0], p1 = _ctrlpts [1],
   p2 = _ctrlpts [2], p3 = _ctrlpts [3],
   p4 = _ctrlpts [4], p5 = _ctrlpts [5],
   p6 = _ctrlpts [6], p7 = _ctrlpts [7],
   p8 = _ctrlpts [8], p9 = _ctrlpts [9],
   p10 = _ctrlpts [10], p11 = _ctrlpts [11],
   p12 = _ctrlpts [12], p13 = _ctrlpts [13],
   p14 = _ctrlpts [14], p15 = _ctrlpts [15];

  // Define G mats
  // Note that the actual matrix layout is not the same as code layout, as
  //   it's column major. So the "look" of it is the transpose of the code look
  Matrix4 Gbez_x = Matrix4 (
    p0[0], p4[0], p8[0], p12[0],
    p1[0], p5[0], p9[0], p13[0],
    p2[0], p6[0], p10[0], p14[0],
    p3[0], p7[0], p11[0], p15[0]);
  Matrix4 Gbez_y = Matrix4 (
    p0[1], p4[1], p8[1], p12[1],
    p1[1], p5[1], p9[1], p13[1],
    p2[1], p6[1], p10[1], p14[1],
    p3[1], p7[1], p11[1], p15[1]);
  Matrix4 Gbez_z = Matrix4 (
    p0[2], p4[2], p8[2], p12[2],
    p1[2], p5[2], p9[2], p13[2],
    p2[2], p6[2], p10[2], p14[2],
    p3[2], p7[2], p11[2], p15[2]);


  // C

  // Full form of C is really:
  // Cx = V * Bbez * Gbez_x * Bbez * U
  // Cy = V * Bbez * Gbez_y * Bbez * U
  // Cz = V * Bbez * Gbez_z * Bbez * U

  // multiply Matrices to get C
  // Cx = Bbez.transpose * Gbez_x * Bbez
  // Cy = Bbez.transpose * Gbez_y * Bbez
  // Cz = Bbez.transpose * Gbez_z * Bbez
  //   Note that Bbez.transpose == Bbez, `.` it's a triangle symmetric mat


  Matrix4 Cx = Matrix4 ();
  Bbez.mul (Gbez_x, Cx);
  Cx.mul (Bbez, Cx);

  Matrix4 Cy = Matrix4 ();
  Bbez.mul (Gbez_y, Cy);
  Cy.mul (Bbez, Cy);

  Matrix4 Cz = Matrix4 ();
  Bbez.mul (Gbez_z, Cz);
  Cz.mul (Bbez, Cz);


  // U, V

  // Parameters for curve at each pt x(u,v)
  float u = 0.0f, v = 0.0f;
  float u_step = 1.0 / _num_samplepts;
  float v_step = 1.0 / _num_samplepts;
  float u2 = 0.0f, u3 = 0.0f, v2 = 0.0f, v3 = 0.0f;

  Vector4 U = Vector4 ();
  Vector4 V = Vector4 ();


  // X, final pt on surface

  // For result pt
  Vector4 X_x = Vector4 ();
  Vector4 X_y = Vector4 ();
  Vector4 X_z = Vector4 ();
  double Xx, Xy, Xz;
  // Result pt
  Vector3 * x;

  int col1start = 0;
  int col1end = _num_samplepts - 1;
  int col2start = col1end + 1;
  int col2end = col1end + _num_samplepts;

  for (int i = 0; i < _num_samplepts; i++)
  {

    u2 = u * u;
    u3 = u2 * u;
    U.setAll (u3, u2, u, 1);

    v = 0.0f;

    for (int j = 0; j < _num_samplepts; j++)
    {
      v2 = v * v;
      v3 = v2 * v;
      V.setAll (v3, v2, v, 1);

      // Update C

      // multiply C by U and V vectors to get x
      // V.transpose * Cx * U
      // V.transpose * Cy * U
      // V.transpose * Cz * U

      Cx.premul (V, X_x);
      Xx = X_x.mul (U);

      Cy.premul (V, X_y);
      Xy = X_y.mul (U);

      Cz.premul (V, X_z);
      Xz = X_z.mul (U);


      // Result Point x
      x = new Vector3 (Xx, Xy, Xz);
      _verts.push_back (x);

fprintf (stderr, "_verts.at (%d): ", _verts.size() - 1);
_verts.at (_verts.size() - 1) -> print ();

      // Tesselation

      if ((i > 0) && (j > 0))
      {
        // |\
        // |_\   CCW. p0 lower left, p1 upper left, p2 lower right
        _triIndices.push_back (col1start + j);
        _triIndices.push_back (col1start + j - 1);
        _triIndices.push_back (col2start + j);
  
        // __
        // \ |
        //  \|  CCW. p0 lower right, p1 upper left, p2 upper right 
        _triIndices.push_back (col2start + j);
        _triIndices.push_back (col1start + j - 1);
        _triIndices.push_back (col2start + j - 1);
      }

      v += v_step;
    }

    u += u_step;

    // Set for next column
    col1start = col2start;
    col1end = col2end;
    col2start = col2end + 1;
    col2end += _num_samplepts;

  }

  fprintf (stderr, "Number of sample pts: %d\n", _verts.size ());
  fprintf (stderr, "Number of triangles: %f\n", _triIndices.size () / 3.0);

  for (int i = 0; i < _triIndices.size (); i += 3)
    fprintf (stderr, "%d %d %d\n", _triIndices.at (i), _triIndices.at(i+1),
      _triIndices.at(i+2));
}





