//
// (Mabel) Mengzi Zhang
// 
// CSE 167 hw 7
// 25 November 2010
//

// OpenGL
#include <GL/glut.h>

// local
#include <shape3D.h>
#include <Matrix4.h>
#include <material.h>

#include <assert.h>
#include "tga1.h"

#define RAD_DEFAULT 1.0
#define HEIGHT_DEFAULT 2.0

// ------------------------------------------------------------------- ctor --
Shape3D::Shape3D ()
{
  _shape_id = SHAPE_CONE;  // Good for debugging and rotating
  //_shape_id = SHAPE_SPHERE;
  init ();
}

Shape3D::Shape3D (Shape * s)
{
  _self = s;
  init ();
}

// Parameters:
//   filename: can be NULL unless id == SHAPE_OBJ, then supply a obj file
//     name.
Shape3D::Shape3D (int id, char * filename)
{
  _shape_id = id;
  // Load obj geometry from file
  if (_shape_id == SHAPE_OBJ)
  {
    objFile = ObjFile ();
    *(_objFilename) = 0;
    //strcat (_objFilename, "in/teapot.obj");
    //strcat (_objFilename, "in/bunny.obj");
    strcat (_objFilename, filename);

    // Load obj file
    Matrix4 transformMat = Matrix4 ();
    // Load obj file, loadObj will set transformMat to center obj file
    objFile.loadObj (minmax, transformMat, _objFilename);

    // Store min max for bd sphere
    calcBdSphere ();

    objFile.getMatrix ().identity ();
    objFile.getMatrix ().mul (transformMat, objFile.getMatrix ());
  }
  init ();
}

void Shape3D::init ()
{
  _radius = RAD_DEFAULT;
  _height = HEIGHT_DEFAULT;
  _isWireframe = false;
  _bdsphere_c = Vector3 ();
  _bdsphere_r = 0;
  _trans_bdsphere_c = Vector3 ();
  _trans_bdsphere_r = 0;

  float * minmax = new float [6];
  //float minmax [6];

  // Node's
  name = new char [name_maxlen];
  *name = 0;
  *className = 0;
  strcat (className, "Shape3D");

  // Blending
  _blendOn = false;

  //env mapping
  _sg_envMapOn = false;
  //GL_REFLECTION_MAP_EXT or GL_NORMAL_MAP_EXT
  mode = GL_REFLECTION_MAP_EXT;
  //GL_CLAMP or GL_REPEAT
  wrap = GL_REPEAT;
  mipmaps = 0;

}


// --------------------------------------------------------------- setters  --

void Shape3D::setRadius (float r)
{
  _radius = r;
}

void Shape3D::setHeight (float h)
{
  _height = h;
}

void Shape3D::setWireframe (bool wf)
{
  _isWireframe = wf;
}

void Shape3D::setBlend (bool b)
{
  _blendOn = b;
}


// -------------------------------------------------------------------- fns --


// Call GL fns to set up and draw geometry in scene
// Note: Do NOT change C!!!! it should be const. Parents still use it to call
//   other children in scene graph.
void Shape3D::draw (Matrix4& C)
{
  // Set MODELVIEW mat

  //glMatrixMode (GL_MODELVIEW);

  glLoadIdentity ();

  Matrix4 transf = Matrix4 ();
  // premultiply on left, same as doing glTranslatef (0, 0, -20);
  // diff from hw5: take out this translation
  //transf.translate (0.0, 0.0, -20.0);
  transf.mul (C, transf);

  glMultMatrixd (transf.getPointer ());


  /*
  // Set lighting

  light.setLitNum (GL_LIGHT0);
  light.setAmbient (0.2, 0.2, 0.2, 1.0);
  //light.setAmbient (0.9, 0.7, 0.0, 1.0);
  //light.setDiffuse (0.0, 1.0, 1.0, 1.0);
  //light.setDiffuse (1.0, 0.9, 0.0, 1.0); // yoke yellow
  light.setDiffuse (0.5, 0.0, 0.9, 1.0);
  light.setSpec (1.0, 1.0, 1.0, 1.0);
  glMultMatrixd (trackball_rotmat.getPointer ());
  //light.setPos (1.0, 2.0, 3.0, 1.0);
  light.setPos (-10.0, 5.0, 3.0, 1.0);
  light.setAttenConst (1.0);
  light.setAttenLin (0.0);
  light.setAttenQuad (0.0);
  */

  /*
  // Set material

  Material material;
  material.reset ();
  //material.setAmbient (0.9, 0.7, 0.0, 1.0);
  material.setAmbient (0.9, 0.0, 0.9, 1.0);
  material.setDiffuse (0.9, 0.9, 0.4, 1.0);
  material.setSpec (0.0, 0.0, 0.0, 1.0);
  material.setShine (0);
  material.setMaterials ();
  */

  // Texture
  if (_sg_textureOn)
  {
    glEnable (GL_TEXTURE_2D);
    // commented out because light was not reflecting off map textures
    //glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
    glBindTexture (GL_TEXTURE_2D, _sg_texName);
  }
  else
    glDisable (GL_TEXTURE_2D);

  // Environment mapping
  if(_sg_envMapOn)
  {
    glEnable(GL_TEXTURE_CUBE_MAP_EXT);
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    glEnable(GL_TEXTURE_GEN_R);
    glBindTexture(GL_TEXTURE_CUBE_MAP_EXT,thisTexture);
  }


  // Blending
  if (_blendOn)
  {
    glEnable (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE);
  }


  // Draw geometry

  switch (_shape_id)
  {
    case SHAPE_CUBE:
      if (_isWireframe)
        glutWireCube (_radius);
      else
        glutSolidCube (_radius);
      break;

    case SHAPE_SPHERE:
      if (_isWireframe)
        glutWireSphere (_radius, 12, 12);
      else
        glutSolidSphere (_radius, 12, 12);
      break;

    case SHAPE_CONE:
      if (_isWireframe)
        glutWireCone (_radius, _height, 12, 12);
      else
        glutSolidCone (_radius, _height, 12, 12);
      break;
    case SHAPE_QUAD:
        draw_Quad(_tex_w, _tex_h);
      break;
    case SHAPE_OBJ:
      // Call objFile's draw() fn to draw it
      if (&objFile)
        objFile.draw ();
      break;

    default:
      if (_isWireframe)
        glutWireSphere (_radius, 12, 12);
      else
        glutSolidSphere (_radius, 12, 12);
      break;
  }

  if(_sg_envMapOn)
  {
    glDisable(GL_TEXTURE_CUBE_MAP_EXT);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glDisable(GL_TEXTURE_GEN_R);
  }


  if (_blendOn)
    glDisable (GL_BLEND);
}

void Shape3D::set_mode(int m)
{
  mode = m;
}
void Shape3D::set_wrap(int w)
{
  wrap = w;
}
void Shape3D::set_mipmaps(int mm)
{
  mipmaps = mm;
}


void Shape3D::set_face_files(char **f)
{
  faceFile[0] = f[0];
  faceFile[1] = f[1];
  faceFile[2] = f[2];
  faceFile[3] = f[3];
  faceFile[4] = f[4];
  faceFile[5] = f[5];
  makeCubeMap();
}

void
Shape3D::loadFace(GLenum target, char *filename)
{
  FILE *file;
  gliGenericImage *image;

  file = fopen(filename, "rb");
  if (file == NULL) 
  {
    printf("cm_demo: could not open \"%s\"\n", filename);
    //exit(1);
  }

  image = gliReadTGA(file, filename);
  fclose(file);

  if (mipmaps) 
  {
    gluBuild2DMipmaps(target, image->components,
      image->width, image->height,
      image->format, GL_UNSIGNED_BYTE, image->pixels);
  } 
  else 
  {
    glTexImage2D(target, 0, image->components,
      image->width, image->height, 0,
      image->format, GL_UNSIGNED_BYTE, image->pixels);
  }
}

void
Shape3D::updateTexgen(void)
{
  assert(mode == GL_NORMAL_MAP_EXT || mode == GL_REFLECTION_MAP_EXT);
  glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, mode);
  glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, mode);
  glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, mode);
}

void
Shape3D::updateWrap(void)
{
  glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_S, wrap);
  glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_T, wrap);
}

void
Shape3D::makeCubeMap(void)
{
  int i;
  glBindTexture(GL_TEXTURE_CUBE_MAP_EXT,thisTexture);
  for (i=0; i<6; i++) 
  {
     loadFace(faceTarget[i], faceFile[i]);
  }
  if (mipmaps) 
  {
    glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_MIN_FILTER,
      GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  } 
  else 
  {
    glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  }
  //glEnable(GL_TEXTURE_CUBE_MAP_EXT);

  updateTexgen();
  updateWrap();

  //glEnable(GL_TEXTURE_GEN_S);
  //glEnable(GL_TEXTURE_GEN_T);
  //glEnable(GL_TEXTURE_GEN_R);
}

// Draw function to draw a quad
void Shape3D::draw_Quad(float w, float h){
    float mid_w = w/2.0;
    float mid_h = h/2.0;
    
    glColor3f(1,1,1);
    glBegin(GL_QUADS);
      glNormal3f(1,1,1);
      if(_sg_textureOn)
        glTexCoord2d(0.0,0.0);
      glVertex3f(-mid_w,-mid_h,0);

      if(_sg_textureOn)
        glTexCoord2d(0.0,1.0);
      glVertex3f(-mid_w,mid_h,0);

      if(_sg_textureOn)
        glTexCoord2d(1.0,1.0);
      glVertex3f(mid_w,mid_h,0);

      if(_sg_textureOn)
        glTexCoord2d(1.0,0.0);
      glVertex3f(mid_w,-mid_h,0);
    glEnd();
}
// Calculate center and radius of bounding sphere around this obj
// This calcs bd sphere WITHOUT any effects from overlaying transformation
//   matrices. To get bd sphere with transform mats, must call updateBdSphere()
//   with that mat.
void Shape3D::calcBdSphere ()
{
  // We don't really have to know where it is in world space.
  // We only need to know how big the shape is in local coordinate sys! So then
  //   we can just get the boundaries by the set radius and height and the
  //   SHAPE_ID to know how far the shape will extend, then compute bd sphere
  //   according to that.

  // Default at origin. Unless there is a translation, in that case, call
  //   updateBdSphere() from outside.
  _bdsphere_c.setAll (0,0,0);

  if (_shape_id == SHAPE_SPHERE)
  {
    _bdsphere_r = _radius;
  }

  else if (_shape_id == SHAPE_CUBE)
  {
    // Longest radius is 1/2 * diagonal of cube. Calc diag of cube
    _bdsphere_r = _radius * sqrt(3) / 2.0;
  }

  else if (_shape_id == SHAPE_CONE)
  {
    _bdsphere_r = max (_radius, _height);
  }

  else if (_shape_id == SHAPE_OBJ)
  {
    
    //_bdsphere_r = (float) obj_cellW / 2.0;
    
    //fprintf (stderr, "%f %f %f %f %f %f\n", minmax[0], minmax[1], minmax[2],
    //  minmax[3], minmax[4], minmax[5]);

    // midpt = (min + max) / 2
    _bdsphere_c.setAll (
      (minmax [ObjReader::MIN_X_IDX] + minmax [ObjReader::MAX_X_IDX]) / 2,
      (minmax [ObjReader::MIN_Y_IDX] + minmax [ObjReader::MAX_Y_IDX]) / 2,
      (minmax [ObjReader::MIN_Z_IDX] + minmax [ObjReader::MAX_Z_IDX]) / 2
    );

    // radius = a farthest pt - cent (either min or max should have same dist
    //   to center, `.` center is set as midpt btwn them. so either way
    //   dist is same.)
    Vector3 res = Vector3 (minmax [ObjReader::MIN_X_IDX],
      minmax [ObjReader::MIN_Y_IDX],
      minmax [ObjReader::MIN_Z_IDX]);
    res.sub (_bdsphere_c, res);
    _bdsphere_r = (float) res.len ();
  }
  

}

float ObjFile::min (float num1, float num2)
{
  if (num1 < num2)
    return num1;
  else
    return num2;
}

// Returns max of the two args
float Shape3D::max (float num1, float num2)
{
  if (num1 > num2)
    return num1;
  else
    return num2;
}


// Update bounding sphere with transform matrix applied to the shape.
//   Center updated just by multiplying mat by center Vec3.
//   Radius updated by creating a pt on sphere (ANY pt that's _bdsphere_r
//   dist from center. Very easy), multiply that pt by 
// Parameter:
//   transform: matrix that will be applied to this shape when drawn in 
//     frame buffer.
void Shape3D::updateBdSphere (Matrix4& transf)
{
  // Find new bounding sphere after transformation

  // center

  _trans_bdsphere_c.setAll (0,0,0);
  transf.mul (_bdsphere_c, _trans_bdsphere_c);


  // radius

  // Construct a pt _bdsphere_r units from bd sphere center
  Vector3 edge_pt = Vector3 (_bdsphere_c[0] + _bdsphere_r, _bdsphere_c[1],
    _bdsphere_c[2]);

  // Transform this pt with matrix
  transf.mul (edge_pt, edge_pt);

  // Calc difference vec from edge to center = edge_pt - _trans_center
  Vector3 diff = Vector3 ();
  edge_pt.sub (_trans_bdsphere_c, diff);

  // Length of diff vec is new radius
  _trans_bdsphere_r = diff.len ();
}

// Use radius of bounding sphere to calculate whether sphere intersects given
//   plane, represented by a normal and a point on the plane
//   (TODO: see scene management slide p14, Alternatively can represent a plane
//   by just normal and d, where d = dist from origin to plane. If that's
//   easier, use that, instead of pt. But in order to get d, you still need
//   p*n, so I don't see how that's easier.)
// Assumes _bdsphere_c and _bdsphere_r have been computed by calcBdSphere().
// Parameter:
//   normal: normal to the plane
//   pt: a point on the plane
//   //d: distance from origin to plane
// Returns true if sphere is inside plane (on the side of negative normal) or
//   interesecting plane,
//   false if sphere is ENTIRELY outside plane.
// Returns:
//   -1 if sphere ENTIRELY outside plane,
//   0 if sphere intersects plane,
//   1 if sphere ENTIRELY inside plane.
int Shape3D::intersectBdSphere (Vector3& normal, Vector3& pt)
//bool Shape3D::intersectBdSphere (Vector3 normal, float d)
{
  //fprintf (stderr, "intersectBdSphere ().\n");

  // Calc perpendicular distance of bounding sphere's center to given plane
  // dist = (x - p) dot n, i.e. get vector from pt p on plane to pt x in space,
  //   then project this vector onto normal n by dot product with n.
  //   This will get the perpendicular distance of pt x to the plane.
  // This vec will have a SIGN, resulting in distance having a sign.
  //   positive: sphere is outside plane.
  //     (`.` it means sphere is on the side of normal of plane, since normal
  //     is always on outside of plane so you can see it, sphere is outside
  //     plane.)
  //   negative: sphere is inside plane.
  Vector3 vec_xp = Vector3 ();
//fprintf (stderr, "shape3D::draw(): pt %f %f %f\n", pt[0], pt[1], pt[2]);
//fprintf (stderr, "shape3D::draw(): normal %f %f %f\n", normal[0], normal[1], normal[2]);
  _trans_bdsphere_c.sub (pt, vec_xp);
  double dist = vec_xp.dot (normal);


  // Completely outside plane
  if (dist > _trans_bdsphere_r)
    return -1;
  // Completely inside plane
  else if (dist < -_trans_bdsphere_r)
    return 1;
  // Intersects plane
  //if ((-_trans_bdsphere_r < dist) && (dist < _trans_bdsphere_r))
  else
    return 0;


  /*
  // If dist is positive, meaning center of bding sphere is outside of plane,
  //   then test if rest of sphere might be intersecting plane.
  if (dist > 0.0)
  {
    //fprintf (stderr, "dist > 0\n");

    // If positive dist is greater than positive radius, i.e.
    //   dist from center of sphere to plane is greater than
    //   dist from center of sphere to edge of sphere (i.e. radius),
    //   meaning that the plane is completely outside the sphere.
    // Assume _bdsphere_r is always positive, `.` it's a radius we computed.
    if (dist > _trans_bdsphere_r)
      return -1;

    // Sphere is intersecting plane, for at least 1 point on the tangent even
    //   if all remaining sphere is outside plane.
    else
      return 0;
  }

  // If dist negative, meaning center of bding sphere is inside plane
  else if (dist < 0.0)
  {
    //fprintf (stderr, "dist < 0\n");

    // If negative dist is smaller than negative radius, (i.e. think of abs
    //   value, abs (dist) > abs (_bdsphere_r) ), then entire sphere is
    //   completely inside plane.
    // Assume _bdsphere_r is always positive, `.` it's a radius we computed.
    if (dist < -_trans_bdsphere_r)
      return 1;

    // Only a part of the sphere is inside plane
    else
      return 0;
  }

  // (dist == 0), sphere's center pt is exactly on the plane
  else
  {
    //fprintf (stderr, "dist == 0\n");
    return 0;
  }
  */

}




// Obj files code from hw2
// -------------------------------------------------------------- Obj files --
  
ObjFile::ObjFile ()
{ 
  Shape (); 
} 

  
// Call routine to load and read .obj file.
// Parameters:
//   transformMat: matrix to store the final transformation 
void ObjFile::loadObj (float * minmax,
  Matrix4& transformMat, char * objFilename)
{
  fprintf (stderr, "Reading obj file %s... ", objFilename);
  
  // To store min max x y z for scale and translation to fit unit cube window
  //float minmax [6];

  // Read an obj file
  ObjReader::readObj (objFilename, nVerts, &verts, &normals, &texcoords,
    nIndices, &indices, minmax);

  fprintf (stderr, "Loaded.\n");
  
  normalize (minmax, transformMat);
}
  
  
// Calculate translation and scale matrices so that the obj model can be
//   centered at (0,0,0) and fits in the unit cube.
void ObjFile::normalize (float * minmax, Matrix4& transformMat)
{
  float min_x = minmax [ObjReader::MIN_X_IDX];
  float max_x = minmax [ObjReader::MAX_X_IDX];
  float min_y = minmax [ObjReader::MIN_Y_IDX];
  float max_y = minmax [ObjReader::MAX_Y_IDX];
  float min_z = minmax [ObjReader::MIN_Z_IDX];
  float max_z = minmax [ObjReader::MAX_Z_IDX];
  

  // Calc translation matrix for .obj model

  float x_trans = 0, y_trans = 0, z_trans = 0;

  // Translate to make object fit into unit cube
  // Calculate midpoint. If midpoint is not at (0,0,0), translate the inverse
  //   i.e. negate, to make the midpoint at (0,0,0)
  x_trans = - ((max_x + min_x) / 2);
  y_trans = - ((max_y + min_y) / 2);
  z_trans = - ((max_z + min_z) / 2);

  Matrix4 transMat = Matrix4 ();
  transMat.translate (x_trans, y_trans, z_trans);


  // Calc scale matrix for .obj model

  // Min ratio of window size to object size, used for uniform scaling (thus
  //   need min. Else might scale too much if use max)
  float factor = min (obj_cellW / (float) (max_x - min_x),
    obj_cellH / (float) (max_y - min_y));

  // Debug
  //fprintf (stderr, "x_width %f y_height %f scaling factor %f",
  //  (float) max_x - min_x, (float) max_y - min_y, factor);

  //factor /= 10;
  //factor = pow (factor, (float) 1/3);
  factor = pow (factor, (float) 1/2);

  Matrix4 scaleMat = Matrix4 ();
  scaleMat.scale (factor, factor, factor);


  scaleMat.mul (transMat, transformMat);
}


// Draw shapes read and stored from the obj file
// Note: texcoords data is currently ignored.
void ObjFile::draw ()
{
  float solidColor [] = {1, 1, 0};

  doDraw (nVerts, verts, nIndices, indices, NULL, solidColor, normals, texcoords);
}


