//
// (Mabel) Mengzi Zhang
// 24 Oct 2010
// CSE 167 hw 4
//
// Vertex shader program for spot light source.
// Effects applied on gl_LightSource [1], i.e. GL_LIGHT1.
// Calculates the light direction to this vertex, `.` pt light emits in all
//   dirs, so each vert has a diff direction vec with light source.
// Calculates distance from light source to point and passes to fragment prgm
//   where attenuation is calculated (can't just calc attenuation here and
//   pass to fragment prgm to interpolate `.` attenuation is not linear).
//

// Per-vertex shader program. Applied on each vertex of a triangle.

// Vertex shader for per-pixel diffuse shading.The teapot is a good object to
//   test this shader, since the .obj file includes surface normals.

// The shader computes the unit light direction and unit surface normal, which
//   are passed to the fragment shader as varying variables.


// Outputs from ths vertex program, inputs to fragment program
// To use in calculating color (including attenuation) using formula:
//   color = ambientGlobal + attenuation (ambient + diffuse + specular)
//   where attenuation = 1 / (k0 + k1d + k2d^2)
varying vec4 diffuse, ambientGlobal, ambient;
varying vec3 normal, lightDir, halfVector;
varying float dist;

void main()
{
  vec4 ecPos;
  vec3 aux;

  // Compute user-defined varying var to pass to fragment prgm
  // gl_LightSource, gl_NormalMatrix, and gl_Normal are pre-defined vars that
  //   access the current OpenGL state.
  normal = normalize (gl_NormalMatrix * gl_Normal);

  // Compute user-defined varying vars to pass to fragment program
  // Computer light's direction
  //lightDir = normalize (vec3 (gl_LightSource [1].position));
  ecPos = gl_ModelViewMatrix * gl_Vertex;
  aux = vec3 (gl_LightSource [1].position - ecPos);
  lightDir = normalize (aux);
  dist = length (aux);
  
  halfVector = normalize (gl_LightSource [1].halfVector.xyz);

  // Compute diffuse, ambient, ambientGlobal
  // Calc diffuse from obj's diffuse color and light's diffuse color
  diffuse = gl_FrontMaterial.diffuse * gl_LightSource [1].diffuse;
  // Ambient terms are separated `.` local ambient needs to go through
  //   attenuation separately
  // Calc ambient from obj's ambient color and light's ambient color
  ambient = gl_FrontMaterial.ambient * gl_LightSource [1].ambient;
  // Calc global ambient from global light model's ambient & obj's ambient
  ambientGlobal = gl_LightModel.ambient * gl_FrontMaterial.ambient;

  // Required output.
  // ftransform() is a built-in function that applies all transformations
  //   (i.e., modelview and projection) to a vertex.
  gl_Position = ftransform();
}

